import txaio

from pysnmp.hlapi.twisted import getCmd, SnmpEngine, CommunityData, UdpTransportTarget,\
                                 ContextData, ObjectType, ObjectIdentity

# For logging
txaio.use_twisted()


class SNMPTwister:
    """Helper class for handling SNMP communication with twisted.

    More information can be found in the pySNMP documentation. The
    `SNMP Operations`_ page is particularly helpful for understanding the setup
    of this object.

    Note: This helper currently only supports SNMPv1.

    .. _SNMP Operations: http://snmplabs.com/pysnmp/docs/pysnmp-hlapi-tutorial.html

    Parameters
    ----------
    address : str
        Address of the SNMP Agent to send GET/SET requests to
    port : int
        Associated port for SNMP communication. Default is 161

    Attributes
    ----------
    snmp_engine : pysnmp.entity.engine.SnmpEngine
        PySNMP engine
    address : str
        Address of the SNMP Agent to send GET/SET requests to
    udp_transport : pysnmp.hlapi.twisted.transport.UdpTransportTarget
        UDP transport for UDP over IPv4
    log : txaio.tx.Logger
        txaio logger object

    """

    def __init__(self, address, port=161):
        self.snmp_engine = SnmpEngine()
        self.address = address
        self.udp_transport = UdpTransportTarget((address, port))
        self.log = txaio.make_logger()

    def _success(self, args):
        """Success callback for getCmd.

        Taken from Twisted example for SNMPv1 from pySNMP documentation:
        http://snmplabs.com/pysnmp/examples/hlapi/twisted/contents.html

        Returns
        -------
        list
            A sequence of ObjectType class instances representing MIB variables
            returned in SNMP response.

        """
        (error_status, error_index, var_binds) = args

        if error_status:
            self.log.error('%s: %s at %s' % (self.address,
                                             error_status.prettyPrint(),
                                             error_index and
                                             var_binds[int(error_index) - 1][0] or '?'))
        else:
            for var in var_binds:
                self.log.debug(' = '.join([x.prettyPrint() for x in var]))

        return var_binds

    def _failure(self, error_indication):
        """Failure Errback for getCmd.

        Taken from Twisted example for SNMPv1 from pySNMP documentation:
        http://snmplabs.com/pysnmp/examples/hlapi/twisted/contents.html

        """
        self.log.error('%s failure: %s' % (self.address, error_indication))

    def get(self, oid_list):
        """Issue a getCmd to get SNMP OID states.

        Example
        -------
        >>> snmp = SNMPTwister('localhost', 161)
        >>> snmp.get([ObjectType(ObjectIdentity('MBG-SNMP-LTNG-MIB',
                                                'mbgLtNgRefclockState',
                                                1)),
                      ObjectType(ObjectIdentity('MBG-SNMP-LTNG-MIB',
                                                'mbgLtNgRefclockLeapSecondDate',
                                                1))])

        >>> snmp = SNMPTwister('localhost', 161)
        >>> result = snmp.get([('MBG-SNMP-LTNG-MIB', 'mbgLtNgRefclockState', 1),
                               ('MBG-SNMP-LTNG-MIB', 'mbgLtNgRefclockLeapSecondDate', 1)])
        >>> # Simply printing the returned object shows a nice string
        >>> print(result[0])
        MBG-SNMP-LTNG-MIB::mbgLtNgRefclockState.1 = notSynchronized
        >>> # The corresponding integer value is hidden within the returned object
        >>> print(result[0][1]._value)
        2

        Parameters
        ----------
        oid_list : list
            List of high-level MIB Object OIDs. The list elements should either be
            ObjectType, or tuples which define the OIDs, as shown in the
            example above. See `Specifying MIB Objects`_ for more info.

            .. _Specifying MIB Objects:
               http://snmplabs.com/pysnmp/docs/pysnmp-hlapi-tutorial.html#specifying-mib-object

        Returns
        ------
        twisted.internet.defer.Deferred
            A Deferred which will callback with the var_binds list from
            self._success. If successful, this will contain a list of ObjectType class
            instances representing MIB variables returned in SNMP response.

        """
        oid_list = [ObjectType(ObjectIdentity(*x)) if isinstance(x, tuple) else x for x in oid_list]

        datagram = getCmd(self.snmp_engine,
                          CommunityData('public', mpModel=0),  # SNMPv1
                          self.udp_transport,
                          ContextData(),
                          *oid_list)

        datagram.addCallback(self._success).addErrback(self._failure)

        return datagram



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Feb 23 14:15:51 2018

@author: jacoblashner
"""

from serial import Serial
from serial.serialutil import SerialException
import time
import sys
import numpy as np
from collections import OrderedDict
import socket
from typing import List


BUFF_SIZE = 1024

try:
    from tqdm import *
except ModuleNotFoundError:
    tqdm = lambda x: x

class Module:
    """
        Allows communication to Lakeshore Module.
        Contains list of inputs which can be read from.
    """

    def __init__(self, port='/dev/tty.SLAB_USBtoUART', baud=115200, timeout=10):
        """
            Establish Serial communication and initialize channels.
        """

        # Running with a simulator
        # Make sure to write over tcp instead of serial.
        if port[:6] == 'tcp://':
            self.simulator = True
            address, socket_port = port[6:].split(':')
            socket_port = int(socket_port)

            for p in range(socket_port, socket_port + 10):
                try:
                    print(f"Trying to connect to {address} on port {p}")
                    self.com = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                    self.com.connect((address, p))
                    print(f"Found connection on port {p}")
                    break
                except ConnectionRefusedError as e:
                    if e.errno == 61:
                        continue
                    else:
                        raise e

        else:
            self.com = Serial(port=port, baudrate=baud, timeout=timeout)
            self.simulator = False

        # First comms usually fails if this is your first time communicating
        # after plugging in the LS240. Try three times, then give up.
        for i in range(3):
            try:
                print('attempt %s'%i)
                idn = self.msg("*IDN?")
                break
            except TimeoutError:
                print("Comms failed on attempt %s"%i)

        self.manufacturer, self.model, self.inst_sn, self.firmware_version = idn.split(',')
        num_channels = int(self.model[-2])

        self.name = self.msg("MODNAME?")

        self.channels: List[Channel] = []
        for i in range(num_channels):
            c = Channel(self, i+1)
            self.channels.append(c)

    def close(self):
        if self.simulator:
            self.com.close()

    def __exit__(self):
        self.close()

    def msg(self, msg):
        """
            Send command or query to module.
            Return response (within timeout) if message is a query.
        """
        if self.simulator:
            message_string = "{};".format(msg)
            self.com.send(message_string.encode())
            resp = ''
            if '?' in msg:
                resp = self.com.recv(BUFF_SIZE).decode()
            return resp

        else:
            # Writes message
            message_string = "{}\r\n;".format(msg).encode()

            # write(message_string)
            self.com.write(message_string)

            # Reads response if queried
            resp = ''
            if "?" in msg:
                resp = self.com.readline()
                resp = str(resp[:-2], 'utf-8')       # Strips terminating chars
                if not resp:
                    raise TimeoutError("Device timed out")

            # Must wait 10 ms before sending another command
            time.sleep(.01)

            return resp

    def set_name(self, name):
        self.name = name
        self.msg("MODNAME {}".format(name))

    def __str__(self):
        return "{} ({})".format(self.name, self.inst_sn)


# ==============================================================================
# Lots of stuff to convert between integers that are read by the module
# and what the integers actually stand for
# ==============================================================================

# To convert from int representation to string
sensorStrings = ["None","Diode", "PlatRTC", "NTCRTC"]
unitStrings = ["None", "Kelvin", "Celsius", "Sensor", "Fahrenheit"]

# To convert from int representation to excitation or range
# use:   ranges[sensorType][range]
excitations = [[10e-6], [1e-3], [1e-3, 300e-6, 100e-6, 30e-6, 10e-6, 3e-6, 1e-6, 300e-9, 100e-9]]
ranges = [[7.5], [1e3], [10, 30, 100, 300, 1e3, 3e3, 10e3, 30e3, 100e3]]

units_key = {1: 'K', 2: 'C', 3: 'S', 4: 'F'}


class Channel:
    """Object for each channel of the lakeshore module

    Attributes:
        channel_num (int): The number of the channel (1-8). This should not be changed once set
        name (str): Specifies name of channel
        sensor (int): 1 = Diode, 2 = PlatRTC, 3 = NTC RTD
        auto_range (int): Specifies if channel should use autorange (1,0).
        range (int): Specifies range if auto_range is false (0-8). Range is accoriding to Lakeshore docs.
        current_reversal (int): Specifies if current reversal should be used (0, 1). Should be 0 for diode.
        unit (int): 1 = K, 2 = C, 3 = Sensor, 4 = F
        enabled (int): Sets whether channel is enabled. (1,0)

    """
    def __init__(self, ls, channel_num):
        self.ls = ls
        self.channel_num = channel_num

        # Reads channel info from device
        response = self.ls.msg("INTYPE? {}".format(self.channel_num))
        data = response.split(',')

        self._sensor = int(data[0])
        self._auto_range = int(data[1])
        self._range = int(data[2])
        self._current_reversal = int(data[3])
        self._unit = int(data[4])
        self._enabled = int(data[5])

        response = self.ls.msg("INNAME? %d" % (self.channel_num))
        self.name = response.strip()

    def set_values(self, sensor=None, auto_range=None, range=None,
                   current_reversal=None, unit=None, enabled=None, name=None):
        """Sets Channel parameters after validation.

        Args:
            channel (int):
                Channel number to set. Valid choices are 1-8.
            sensor (int, optional):
                Specifies sensor type:

                    +---+---------+
                    | 1 | Diode   |
                    +---+---------+
                    | 2 | PlatRTC |
                    +---+---------+
                    | 3 | NTC RTD |
                    +---+---------+

            auto_range (int, optional):
                Specifies if channel should use autorange. Must be 0 or 1.
            range (int, optional):
                Specifies range if auto_range is false. Only settable for NTC
                RTD:

                    +---+--------------------+
                    | 0 | 10 Ohms (1 mA)     |
                    +---+--------------------+
                    | 1 | 30 Ohms (300 uA)   |
                    +---+--------------------+
                    | 2 | 100 Ohms (100 uA)  |
                    +---+--------------------+
                    | 3 | 300 Ohms (30 uA)   |
                    +---+--------------------+
                    | 4 | 1 kOhm (10 uA)     |
                    +---+--------------------+
                    | 5 | 3 kOhms (3 uA)     |
                    +---+--------------------+
                    | 6 | 10 kOhms (1 uA)    |
                    +---+--------------------+
                    | 7 | 30 kOhms (300 nA)  |
                    +---+--------------------+
                    | 8 | 100 kOhms (100 nA) |
                    +---+--------------------+

            current_reversal (int, optional):
                Specifies if input current reversal is on or off.
                Always 0 if input is a diode.
            units (int, optional):
                Specifies preferred units parameter, and sets the units for
                alarm settings:

                    +---+------------+
                    | 1 | Kelvin     |
                    +---+------------+
                    | 2 | Celsius    |
                    +---+------------+
                    | 3 | Sensor     |
                    +---+------------+
                    | 4 | Fahrenheit |
                    +---+------------+

            enabled (int, optional):
                Sets if channel is enabled.
            name (str, optional):
                Sets name of channel.

        """
        # Checks to see if values are valid
        if sensor is not None:
            if sensor in [1, 2]:
                self._sensor = sensor
                self._range = 0
            elif sensor == 3:
                self._sensor = sensor
            else:
                print("Sensor value must be 1,2, or 3.")

        if auto_range is not None:
            if auto_range in [0, 1]:
                self._auto_range = auto_range
            else:
                print("auto_range must be 0 or 1.")

        if range is not None:
            if self._sensor == 3 and range in [0, 1, 2, 3, 4, 5, 6, 7, 8]:
                self._range = range
            elif range == 0:
                self._range = range
            else:
                print("Range must be 0 for Diode or Plat RTD, or 0-8 for a NTC RTD")

        if current_reversal is not None:
            if current_reversal in [0, 1]:
                self._current_reversal = current_reversal
            else:
                print("current_reversal must be 0 or 1.")

        if unit is not None:
            if unit in [1, 2, 3, 4]:
                self._unit = unit
            else:
                print("unit must be 1, 2, 3, or 4")

        if enabled is not None:
            if enabled in [0, 1]:
                self._enabled = enabled
            else:
                print("enabled must be 0 or 1")

        if name is not None:
            self.name = name

        # Writes new values to module
        self.ls.msg("INNAME {},{!s}".format(self.channel_num, self.name))

        input_type_message = "INTYPE "
        input_type_message += ",".join(["{}".format(c) for c in [ self.channel_num, self._sensor, self._auto_range,
                                                                    self._range, self._current_reversal, self._unit,
                                                                    int(self._enabled)]])
        self.ls.msg(input_type_message)

    def read_curve(self):
        # Reads curve
        breakpoints = []
        for i in range(1, 201):
            resp = self.ls.msg("CRVPT? {},{}".format(self.channel_num, i))
            unit, temp = resp.split(',')
            if float(unit) == 0.0:
                break
            breakpoints.append((float(unit), float(temp)))

        resp = self.ls.msg("CRVHDR? {}".format(self.channel_num)).split(',')

        header = {
            "Sensor Model": resp[0],
            "Serial Number": resp[1],
            "Data Format": int(resp[2]),
            "SetPoint Limit": float(resp[3]),
            "Temperature Coefficient": int(resp[4]),
            "Number of Breakpoints": len(breakpoints)
        }

        self.curve = Curve(header=header, breakpoints=breakpoints)

    def get_reading(self, unit=None):
        """Get a reading from the channel in the specified units.

        If no unit is provided, use the one determined by the channel settings.

        Args:
            unit (str): Units for reading, options are Kelvin (K), Celcius (C),
                        Fahrenheit (F), or Sensor (S)

        """
        if unit is None:
            u = units_key[self._unit]
        else:
            u = unit

        assert u.upper() in ['K', 'C', 'F', 'S']

        message = "{}RDG? {}".format(u, self.channel_num)
        response = self.ls.msg(message)

        return float(response)

    def load_curve_point(self, n, x, y):
        """ Loads point n in the curve for specified channel"""
        message = "CRVPT "
        message += ",".join([str(c) for c in [self.channel_num, n, x, y]])
        self.ls.msg(message)

    def load_curve(self, filename):
        """Upload calibration curve to channel from file.

        Args:
            filename (str): Calibration file for upload.

        """
        self.curve = Curve(filename=filename)
        hdr = self.curve.header
        keys = list(hdr)

        #loads header
        cmd = "CRVHDR {}".format(self.channel_num)
        for key in keys[:5]:
            cmd += ",{}".format(hdr[key])
        print(cmd)
        self.ls.msg(cmd)

        bps = self.curve.breakpoints
        assert len(bps) <= 200, "Curve must have 200 breakpoints or less"

        print ("Loading Curve to {}".format(self.name))
        for i in range(200):
            if i < len(bps):
                self.load_curve_point(i+1, bps[i][0], bps[i][1])
            else:
                self.load_curve_point(i+1, 0, 0)
        print("Curve loaded")

    def delete_curve(self):
        """Delete calibration curve from channel."""
        cmd = "CRVDEL {}".format(self.channel_num)
        self.ls.msg(cmd)

    def __str__(self):
        string = "-" * 40 + "\n"
        string += "{} -- Channel {}: {}\n".format(self.ls.inst_sn, self.channel_num, self.name)
        string += "-"*40 + "\n"

        string += "{!s:<18} {!s:>13}\n".format("Enabled:", self._enabled)
        string += "{!s:<18} {!s:>13} ({})\n".format("Sensor:", self._sensor, sensorStrings[self._sensor])
        string += "{!s:<18} {!s:>13}\n".format("Auto Range:", self._auto_range)

        range_unit = "V" if self._sensor == 1 else "Ohm"
        string += "{!s:<18} {!s:>13} ({} {})\n".format("Range:", self._range, ranges[self._sensor-1][self._range], range_unit)
        string += "{!s:<18} {!s:>13}\n".format("Current Reversal:", self._current_reversal)
        string += "{!s:<18} {!s:>13}\n".format("Units:", units_key[self._unit])

        return string


class Curve:
    """
    Header for calibration curve::

        :Sensor Model:      Name of curve
        :Serial Number:        Serial Number
        :Data Format:    2 = V:K, 3 = Ohms:K, 4 = log(Ohms):K
        :SetPoint Limit:     Temperature Limit (in K)
        :Temperature Coefficient:     1 = negative, 2 = positive
        :Number of Breakpoints:     Number of curve points

    """
    def __init__(self, filename=None, header=None, breakpoints=None):

        if filename is not None:
            self.load_from_file(filename)
        else:
            if header and breakpoints:
                self.header = header
                self.breakpoints = breakpoints
            else:
                raise Exception("Must give either filename or header and breakpoints")

    def write_to_file(self, filename):
        with open(filename, 'w') as file:

            keys = list(self.header)
            for k in keys:
                print(k, self.header[k])
                file.write("{}:\t{}\n".format(k, self.header[k]))

            file.write('\n')
            file.write('No.\tUnits\tTemperature (K)\n')
            file.write('\n')

            for i,bp in enumerate(self.breakpoints):
                file.write('{}\t{:.4f} {:.4f}\n'.format(i+1, bp[0], bp[1]))

    def load_from_file(self, filename):
        with open(filename, 'r') as file:
            content = file.readlines()

        self.header = OrderedDict({})
        for line in content:
            if line.strip()=='':
                break
            key, v = line.split(':')
            val = v.split('(')[0].strip()
            self.header[key] = val

        self.breakpoints = []
        for line in content[9:]:
            num, unit, temp = line.split()
            self.breakpoints.append((float(unit), float(temp)))

    def __str__(self):
        string = ""
        for key, val in self.header.items():
            string += "%-15s: %s\n"%(key, val)
        return string

if __name__ == "__main__":
    ls = Module(port=sys.argv[1])
    print (ls)



import hashlib
from typing import ContextManager
from contextlib import contextmanager
import os


def get_md5sum(filename):
    m = hashlib.md5()

    for line in open(filename, 'rb'):
        m.update(line)
    return m.hexdigest()


@contextmanager
def get_db_connection(**config):
    """
    Mysql connection context manager.

    Same args as mysql.connector:
    https://dev.mysql.com/doc/connector-python/en/connector-python-connectargs.html
    """
    import mysql.connector
    con = mysql.connector.connect(**config)
    try:
        yield con
    finally:
        con.close()



import os
from enum import Enum

import time
import txaio
import numpy as np
import socket

# For logging
txaio.use_twisted()

on_rtd = os.environ.get('READTHEDOCS') == 'True'
if not on_rtd:
    import so3g
    from spt3g import core


class FlowControl(Enum):
    """Flow control enumeration."""
    ALIVE = 0
    START = 1
    END = 2
    CLEANSE = 3


def _create_dirname(start_time, data_dir, stream_id):
    """Create the file path for .g3 file output.

    Note: This will create directories if they don't exist already.

    Parameters
    ----------
    start_time : float
        Timestamp for start of data collection
    data_dir : str
        Top level data directory for output
    stream_id : str
        Stream id of collection

    Returns
    -------
    str
        dirname for file for output

    """
    sub_dir = os.path.join(data_dir,
                           "{:.5}".format(str(start_time)),
                           stream_id)

    # Create new dir for current day
    if not os.path.exists(sub_dir):
        os.makedirs(sub_dir)

    return sub_dir


def _create_file_path(dirname, basename, suffix="", extension="g3"):
    """Create the full path to the output file for writing.

    Parameters
    ----------
    dirname : str
        dirname for output, likely from _create_dirname
    basename : str
        basename of file without extension or suffix
    suffix : str
        Suffix to append to filename (include your own '_')
    extension : str
        file extension to put on file (without '.')

    Returns
    -------
    str
        Full filepath of file for output

    """
    filename = "{}{}.{}".format(basename, suffix, extension)
    filepath = os.path.join(dirname, filename)

    return filepath


class FrameRecorder:
    """Frame recording object. Keeps track of the reader/writer and associated
    information, such as the frames being written to file and the location of
    the output files.

    All of the methods were separated from the loop that ran in the
    TimestreamAggregator OCS Agent. They're meant to be looped over still and
    for that we provide the run() method.

    This class is not safe to run in the twisted reactor, run in a worker
    thread.

    Parameters
    ----------
    file_duration : int
        seconds per file
    tcp_addr : str
        tcp connection address (i.e. "tcp://127.0.0.1:4536")
    data_dir : str
        Location to write data files to. Subdirectories will be created within
        this directory roughly corresponding to one day.
    stream_id : str
        Stream ID to use to determine file path if one is not found in frames.

    Attributes
    ----------
    time_per_file : int
        Amount of time in seconds to have in each file. Files will be rotated
        after this many seconds.
    address : str
        Address of the host sending the data via the G3NetworkSender.
    data_dir : str
        Location to write data files to. Subdirectories will be created within
        this directory roughly corresponding to one day.
    log : txaio.tx.Logger
        txaio logger object.
    reader : spt3g.core.G3Reader
        G3Reader object to read the frames from the G3NetworkSender.
    writer : spt3g.core.G3Writer
        G3Writer for writing the frames to disk.
    data_received : bool
        Whether data has been received by the current instance of the G3Reader.
    frames : list
        List of frames that have been read from the network. Gets cleared after
        writing to file.
    last_meta
        The last meta data Frame passed to the G3Writer. The last meta data
        frame gets written to the top of each file.
    last_frame_write_time : float
        The time at which we wrote the last frame. Used to track when files
        should be rotated between acquisitions.
    last_connection_time : float
        The time at which we last attempted to establish a network connection.
        Used to rate limit the number of attempted connections to the
        G3NetworkSender.
    _last_file_timestamp : int
        Timestamp of last filename. Used to check if a new file was started
        within the same second a new file is being made to avoid naming
        conflicts.
    filename_suffix : int
        For enumerating files within an acquisition. We want to split files in
        10 minute intervals, but the names should remain the same with a
        "_0001" style suffix, which we track with this attribute.
    start_time : float
        Start time of current file. Only agrees with basename time if
        filename_suffix == 0.
    dirname : str
        The directory path currently in use.
    basename : str
        The basename of the file currently being used. The actual file will
        also contain the tracked suffix.
    monitored_channels : list
        List of readout channels to be monitored. Data from these channels will
        be stored in ``stream_data`` after each ``run`` function call.
    target_rate : float
        Target sampling rate for monitored channels in Hz.
    stream_data : dict
        Data containing downsampled timestream data from the monitored. This
        dict will have the channel name (such as ``r0012``) as the key, and
        the values will have the structure required by the OCS_Feed.
    """
    def __init__(self, file_duration, tcp_addr, data_dir, stream_id,
                 target_rate=10):
        # Parameters
        self.time_per_file = file_duration
        self.address = tcp_addr
        self.data_dir = data_dir
        self.stream_id = stream_id
        self.log = txaio.make_logger()

        # Reader/Writer
        self.reader = None
        self.writer = None
        self.data_received = False

        # Attributes
        self.frames = []
        self.last_meta = None
        self.last_frame_write_time = None
        self.last_connection_time = None
        self._last_file_timestamp = None

        self.filename_suffix = 0
        self.start_time = None
        self.dirname = None
        self.basename = None

        self.monitored_channels = []
        self.target_rate = target_rate
        self.stream_data = {}

    def __del__(self):
        """Clean up by closing out the file once writing is complete."""
        if self.writer is not None:
            self.close_file()

    def _establish_reader_connection(self, timeout=5):
        """Establish the connection to the G3NetworkSender.

        Attempts to connect once and waits if connection could not be made and
        last connection attempt was very recent.

        Parameters
        ----------
        timeout : int
            Timeout in seconds for the G3Reader, afterwhich the connection will
            drop and G3Reader will return empty lists on each Process call.

        Returns
        -------
        reader : spt3g.core.G3Reader
            The G3Reader object connected to the configured address and port

        """
        reader = None

        try:
            reader = core.G3Reader(self.address,
                                   timeout=timeout)
            self.log.debug("G3Reader connection to {addr} established!",
                           addr=self.address)
        except RuntimeError:
            self.log.error("G3Reader could not connect.")

        # Prevent rapid connection attempts
        if self.last_connection_time is not None:
            t_diff = time.time() - self.last_connection_time
            if t_diff < 1:
                self.log.debug("Last connection was only {d} seconds ago. " +
                               "Sleeping for {t}.", d=t_diff, t=(1 - t_diff))
                time.sleep(1 - t_diff)

        self.last_connection_time = time.time()

        return reader

    def read_frames(self, timeout=5):
        """Establish a connection to the G3NetworkSender if one does not exist,
        then try to read frames from it, discarding any flow control frames.

        Clear internally buffered data and cleanup with call to
        self.close_file() if the reader timesout or has otherwise lost its
        connection.

        Parameters
        ----------
        timeout : int
            Timeout in seconds for establishing the G3Reader connection.

        """
        # Try to connect if we are not already
        if self.reader is None:
            self.reader = self._establish_reader_connection(timeout)

            if self.reader is None:
                return

        # Allows tests to Mock a reader
        if type(self.reader) is core.G3Reader:
            self.frames = self.reader.Process(None)

        if self.frames:
            # Handle flow control frames
            for f in self.frames:
                flow = f.get('sostream_flowcontrol')

                if flow is not None:
                    # START; create_new_file
                    if FlowControl(flow) is FlowControl.START:
                        self.close_file()

                    # END; close_file
                    if FlowControl(flow) == FlowControl.END:
                        self.close_file()
                        self.last_meta = None

                    # ALIVE and CLEANSE; do nothing
                    if FlowControl(flow) is FlowControl.CLEANSE:
                        self.log.debug("I saw a CLEANSE frame.")

            # Discard all flow control frames
            self.frames = [x for x in self.frames if 'sostream_flowcontrol' not in x]
            # Discard Pipeline info frame
            self.frames = [x for x in self.frames
                           if x.type != core.G3FrameType.PipelineInfo]
            if self.frames and not self.data_received:
                self.data_received = True
                self.log.info("Started receiving frames from {addr}",
                              addr=self.address)
            return
        else:
            if self.data_received:
                self.log.info("Could not read frames. Connection " +
                              "timed out, or G3NetworkSender offline. " +
                              "Cleaning up...")
            self.close_file()
            self.data_received = False
            self.reader = None

    def check_for_frame_gap(self, gap_size=5):
        """Check for incoming frame time gap. If frames stop coming in likely
        an acquisition has stopped and we need to rotate files and increment
        our file suffix.

        Rotate files to match data acquisitions (assuming 5 seconds apart or
        more until flow control improved on streaming end).

        Parameters
        ----------
        gap_size : int
            gap size threshold in seconds between frames to check for. Close
            file if exceeded.

        """
        # Don't even try to rotate files if writer not active.
        if self.writer is None:
            self.log.debug("Writer not active, not checking frame gap.")
            return

        if self.last_frame_write_time is not None:
            t_diff = time.time() - self.last_frame_write_time
            if t_diff > gap_size:
                self.log.debug("Last frame written more than {g} seconds " +
                               "ago, rotating file", g=t_diff)
                self.close_file()

    def create_new_file(self):
        """Create a new file if needed.

        This is only done if an existing file is closed out (i.e. writer is
        None). Filename and path will be based on the time the file is made.
        Acquisitions that are grouped together by file duration will share the
        same basename as the start of the acquisition and have their suffix
        incremented.

        Writes the last meta data frame to the start of the file if there is
        one.

        Example:
            # Three 10 minute observations
            file_1 = 2019-08-07-01-30-00_000.g3
            file_2 = 2019-08-07-01-30-00_001.g3
            file_3 = 2019-08-07-01-30-00_002.g3

            # Acqusition stopped and new one started on the hour
            new_file = 2019-08-07-02-00-00_000.g3

        """
        if self.writer is None:
            # Used for tracking when to split acquisitions
            self.start_time = time.time()

            # Avoid duplicate filenames if new file started within 1 sec
            if self._last_file_timestamp is None:
                pass
            elif int(self.start_time) == self._last_file_timestamp:
                self.log.debug("New file started within 1 second of previous " +
                               "file, incrementing filename suffix.")
                self.filename_suffix += 1

            # Only create new dir and basename if we've finished an acquisition
            if self.filename_suffix == 0:
                stream_id = self.stream_id
                for f in self.frames:
                    if f.get('sostream_id') is not None:
                        stream_id = f['sostream_id']
                        break

                self.dirname = _create_dirname(self.start_time,
                                               self.data_dir,
                                               stream_id)
                self.basename = int(self.start_time)

            suffix = "_{:03d}".format(self.filename_suffix)
            filepath = _create_file_path(self.dirname, self.basename, suffix)
            self.log.info("Writing to file {}".format(filepath))
            self.writer = core.G3Writer(filename=filepath)
            self._last_file_timestamp = int(self.start_time)

            # Write the last metadata frame to the start of the new file
            if self.last_meta is not None:
                self.writer(self.last_meta)

            return filepath

    def write_frames_to_file(self):
        """Write all frames to file.

        Note: Assumes file writer is already instantiated.

        """
        for f in self.frames:
            # Make sure we do not record flowcontrol frames
            flow = f.get('sostream_flowcontrol')
            if flow is not None:
                self.log.warn("Received flow control frame with value {v}. " +
                              "Flow control frames should be discarded " +
                              "earlier than this", v=flow)
                continue

            # Write most recent meta data frame
            if f.type == core.G3FrameType.Observation:
                self.last_meta = f

            self.writer(f)
            self.writer.Flush()
            self.last_frame_write_time = time.time()

        # clear frames list
        self.frames = []

    def split_acquisition(self):
        """Split the acquisition into multiple files based on file_duration.
        Max file length set by file_duration parameter on FrameRecorder object.

        Only rotates if writer currently active.

        """
        if self.writer is None:
            return

        t_diff = time.time() - self.start_time
        if t_diff > self.time_per_file:
            self.log.debug("{s} seconds elapsed since start of file, " +
                           "splitting acquisition", s=t_diff)
            # Flush internal cache and clean-up (no frame written)
            self.writer(core.G3Frame(core.G3FrameType.EndProcessing))
            self.writer = None
            self.filename_suffix += 1

    def close_file(self):
        """Close a file. Clean up with EndProcessing frame, set self.writer to
        None, rezero suffix.

        """
        if self.writer is not None:
            # Flush internal cache and clean-up (no frame written)
            self.writer(core.G3Frame(core.G3FrameType.EndProcessing))
            self.log.debug("File closed.")
        self.writer = None
        self.filename_suffix = 0

    def run(self):
        """Run the DataRecorder, performing all frame collection, file setup,
        file rotation, and writing to file.

        Will only perform actions if frames can be read, with the exception of
        closing the open file if a gap is encountered (i.e. the current
        acquisition has stopped.)

        Meant to be run in a loop.

        """
        self.read_frames()
        self.check_for_frame_gap(10)
        if len(self.monitored_channels) > 0:
            try:
                self.read_stream_data()
            except Exception as e:
                self.log.warn("Exception thrown when reading stream data:\n{e}", e=e)

        if self.frames:
            self.create_new_file()
            self.write_frames_to_file()
            self.split_acquisition()

    def read_stream_data(self):
        """Reads stream data from ``self.frames``, downsamples it, and stores
        it in ``self.stream_data``.
        """
        chan_keys = [f'r{c:04}' for c in self.monitored_channels]
        self.stream_data = {
            k: {'timestamps': [], 'block_name': k, 'data': {k:[]}}
            for k in chan_keys
        }

        for frame in self.frames:
            if frame.type != core.G3FrameType.Scan:
                continue
            ds_factor = (frame['data'].sample_rate/core.G3Units.Hz) \
                        // self.target_rate
            if np.isnan(ds_factor):
                continue
            ds_factor = max(int(ds_factor), 1)
            n_samples = frame['data'].n_samples
            if 1 < n_samples <= ds_factor:
                ds_factor = n_samples - 1
            times = [
                t.time / core.G3Units.s
                for t in frame['data'].times()[::ds_factor]
            ]
            for key in chan_keys:
                data = frame['data'].get(key)
                if data is None:
                    continue
                self.stream_data[key]['timestamps'].extend(times)
                self.stream_data[key]['data'][key].extend(list(data[::ds_factor]))




"""
Script to create, update, or drop the `pysmurf_files` database
"""
import argparse
import mysql.connector
import getpass


class Column:
    def __init__(self, name, type, opts="", fmt="%s"):
        self.name = name
        self.type = type
        self.opts = opts
        self.fmt = fmt

    def __str__(self):
        return " ".join([self.name, self.type, self.opts])

TABLE_VERSION = 1
table = f'pysmurf_files_v{TABLE_VERSION}'
columns = [
    Column("id", "INT", opts="NOT NULL AUTO_INCREMENT PRIMARY KEY"),
    Column("path", "VARCHAR(260)", opts="NOT NULL"),
    Column("action", "VARCHAR(32)"),
    Column("timestamp", "TIMESTAMP"),
    Column("action_timestamp", "INT"),
    Column("format", "VARCHAR(32)"),
    Column("plot", "TINYINT(1)"),
    Column("site", "VARCHAR(32)"),
    Column("pub_id", "VARCHAR(32)"),
    Column("script_id", "VARCHAR(32)"),
    Column("instance_id", "VARCHAR(32)"),
    Column("copied", "TINYINT(1)"),
    Column("failed_copy_attempts", "INT"),
    Column("md5sum", "BINARY(16)", opts="NOT NULL", fmt="UNHEX(%s)"),
    Column("pysmurf_version", "VARCHAR(64)"),
    Column("socs_version", "VARCHAR(64)"),
]
col_dict = {c.name: c for c in columns}


def add_entry(cur, entry):
    """
    Adds entry to table.
    """
    if not set(col_dict.keys()).issuperset(entry.keys()):
        raise RuntimeError(
            "Invalid file entry provided... \n"
            "Keys given: {}\n"
            "Keys allowed: {}".format(entry.keys(), col_dict.keys())
        )

    keys, vals, fmts = [], [], []
    for k in list(entry.keys()):
        if entry[k] is None:
            continue
        keys.append(k)
        vals.append(entry[k])
        fmts.append(col_dict[k].fmt)

    query = "INSERT INTO {} ({}) VALUES ({})"\
            .format(table, ", ".join(keys), ", ".join(fmts))

    cur.execute(query, tuple(vals))


def create_table(cur, update=True):
    """
    Creates new pysmurf_files table from scratch.

    Args:
        cur (MySQL Cursor):
            cursor to files db
        update (optional, bool):
            Add additional columns if existing table is not up to date.
    """

    cur.execute("SHOW TABLES;")
    table_names = [x[0] for x in cur.fetchall()]
    if f'{table}' not in table_names:
        print(f"Creating {table} table...")
        col_strings = [str(c) for c in columns]
        query = "CREATE TABLE {} ({});".format(table, ", ".join(col_strings))

        try:
            cur.execute(query)
            print(f"Created table {table}")
        except mysql.connector.errors.ProgrammingError as e:
            print(e)
    else:
        print(f"Found table {table} and not updating")


def drop_table(cur):
    """
    Drops pysmurf_files.
    """
    try:
        cur.execute(f"DROP TABLE {table};")
        con.commit()
        print(f"{table} dropped.")
    except mysql.connector.errors.ProgrammingError as e:
        print(e)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('cmd', choices=['create', 'update', 'drop', 'test'])
    parser.add_argument('--password', '-p', type=str,
                        help="Password for development database")
    args = parser.parse_args()

    sql_config = {
        'user': 'development',
        'database': 'files',
        'passwd': args.password
    }

    if sql_config['passwd'] is None:
        sql_config['passwd'] = getpass.getpass("Password for development db: ")

    con = mysql.connector.connect(**sql_config)
    cur = con.cursor()
    try:
        if args.cmd == 'test':
            import datetime
            entry = {
                'path':'/data/pysmurf_test/1568779322_fake_tuning4.txt',
                'type': 'fake_tuning',
                'timestamp': datetime.datetime(2019, 9, 18, 4, 2, 2, 188764),
                'plot': 1,
                'format': 'txt',
                'md5sum': '7ac66c0f148de9519b8bd264312c4d64',
                'site': 'observatory',
                'instance_id': 'pysmurf-monitor',
                'copied': 0,
                'failed_copy_attempts': 0,
                'socs_version': '0+untagged.140.g5307c6b.dirty'
            }

            add_entry(cur, entry)
            con.commit()
        if args.cmd == 'create':
            create_table(cur)
            con.commit()
        elif args.cmd == 'update':
            update_columns(cur)
            con.commit()
        elif args.cmd == 'drop':
            while True:
                resp = input("Are you sure you want to drop pysmurf_files? [y/n]: ")
                if resp.lower().strip() == 'y':
                    drop_table(cur)
                    con.commit()
                    break
                elif resp.lower().strip() == 'n':
                    break
                else:
                    print("Could not recognize input")
    finally:
        print("Closing connection")
        con.close()






import os
import time
import subprocess
import tempfile
import txaio

from sqlalchemy import (Column, create_engine, Integer, String, Float, Boolean)
from sqlalchemy.orm import sessionmaker
from sqlalchemy.ext.declarative import declarative_base

from socs.util import get_md5sum

TABLE_VERSION = 0
txaio.use_twisted()


Base = declarative_base()


class SupRsyncFile(Base):
    """
    Files table utilized by the SupRsync agent.

    Attributes
    ----------
        local_path : String
            Absolute path of the local file to be copied
        local_md5sum : String
            locally calculated checksum
        archive_name : String
            Name of the archive, i.e. `timestreams` or `smurf`. Each archive
            is managed by its own SupRsync instance, so they can be copied to
            different base-dirs or hosts.
        remote_path : String
            Path of the file on the remote server relative to the base-dir.
            specified in the SupRsync agent config.
        remote_md5sum : String, optional
            Md5sum calculated on remote machine
        timestamp : Float
            Timestamp that file was added to db
        copied : Float, optional
            Time at which file was transfered
        removed : Float, optional
            Time at which file was removed from local server.
        failed_copy_attempts : Int
            Number of failed copy attempts
        deletable : Bool
            Whether file should be deleted after copying
    """
    __tablename__ = f"supersync_v{TABLE_VERSION}"

    id = Column(Integer, primary_key=True)
    local_path = Column(String, nullable=False)
    local_md5sum = Column(String, nullable=False)
    archive_name = Column(String, nullable=False)
    remote_path = Column(String, nullable=False)
    timestamp = Column(Float, nullable=False)
    remote_md5sum = Column(String)
    copied = Column(Float)
    removed = Column(Float)
    failed_copy_attempts = Column(Integer, default=0)
    deletable = Column(Boolean, default=True)

    def __str__(self):
        excl = ('_sa_adapter', '_sa_instance_state')
        d = {
            k: v for k, v in vars(self).items()
            if not k.startswith('_') and not any(hasattr(v, a) for a in excl)
        }

        s = "SupRsyncFile:\n"
        s += "\n".join([
            f"    {k}: {v}"
            for k, v in d.items()
        ])
        return s


def create_file(local_path, remote_path, archive_name, local_md5sum=None,
                timestamp=None, deletable=True):
    """
    Creates SupRsyncFiles object.

    Args
    ----
        local_path : String or Path
            Absolute path of the local file to be copied
        remote_path : String
            Path of the file on the remote server relative to the base-dir.
            specified in the SupRsync agent config.
        archive_name : String
            Name of the archive, i.e. `timestreams` or `smurf`. Each archive
            is managed by its own SupRsync instance, so they can be copied to
            different base-dirs or hosts.
        local_md5sum : String, optional
            locally calculated checksum. If not specified, will calculate
            md5sum automatically.
        timestamp:
            Timestamp of file. If None is specified, will use the current
            time.
        deletable : bool
            If true, can be deleted by suprsync agent
    """
    local_path = str(local_path)
    remote_path = str(remote_path)

    if local_md5sum is None:
        local_md5sum = get_md5sum(local_path)

    if timestamp is None:
        timestamp = time.time()

    file = SupRsyncFile(
        local_path=local_path, local_md5sum=local_md5sum,
        remote_path=remote_path, archive_name=archive_name,
        timestamp=timestamp
    )

    if deletable is not None:
        file.deletable = deletable

    return file


class SupRsyncFilesManager:
    """
    Helper class for accessing and adding entries to the SupRsync
    files database.

    Args
    -----
        db_path : path
            path to sqlite db
        create_all : bool
            Create table if it hasn't been generated yet.
        echo : bool
            If true, writes sql statements to stdout
    """
    def __init__(self, db_path, create_all=True, echo=False):
        db_path = os.path.abspath(db_path)
        if not os.path.exists(os.path.dirname(db_path)):
            os.makedirs(os.path.dirname(db_path))

        self._engine = create_engine(f'sqlite:///{db_path}', echo=echo)
        self.Session = sessionmaker(bind=self._engine)

        if create_all:
            Base.metadata.create_all(self._engine)

    def add_file(self, local_path, remote_path, archive_name,
                 local_md5sum=None, timestamp=None, session=None,
                 deletable=True):
        """
        Adds file to the SupRsyncFiles table.

        Args
        ----
            local_path : String
                Absolute path of the local file to be copied
            remote_path : String
                Path of the file on the remote server relative to the base-dir.
                specified in the SupRsync agent config.
            archive_name : String
                Name of the archive, i.e. `timestreams` or `smurf`. Each archive
                is managed by its own SupRsync instance, so they can be copied to
                different base-dirs or hosts.
            local_md5sum : String, optional
                locally calculated checksum. If not specified, will calculate
                md5sum automatically.
            session : sqlalchemy session
                Session to use to add the SupRsyncFile. If None, will create
                a new session and commit afterwards.
            deletable : bool
                If true, can be deleted by suprsync agent
        """
        file = create_file(local_path, remote_path, archive_name,
                           local_md5sum=local_md5sum, timestamp=timestamp,
                           deletable=deletable)
        if session is None:
            with self.Session.begin() as session:
                session.add(file)
        else:
            session.add(file)

    def get_copyable_files(self, archive_name, session=None,
                           max_copy_attempts=None, num_files=None):
        """
        Gets all SupRsyncFiles that are copyable, meaning they satisfy:
         - local and remote md5sums do not match
         - Failed copy attempts is below the max number of attempts

        Args
        ----
            archive_name : string
                Name of archive to get files from
            session : sqlalchemy session
                Session to use to get files. If none is specified, one will
                be created. You need to specify this if you wish to change
                file data and commit afterwards.
            max_copy_attempts : int
                Max number of failed copy atempts
            num_files : int
                Number of files to return
        """
        if session is None:
            session = self.Session()

        query = session.query(SupRsyncFile).filter(
            SupRsyncFile.removed == None,
            SupRsyncFile.archive_name == archive_name,
        )

        if max_copy_attempts is not None:
            query.filter(SupRsyncFile.failed_copy_attempts < max_copy_attempts)

        files = []
        for f in query.all():
            if f.local_md5sum != f.remote_md5sum:
                files.append(f)
            if num_files is not None:
                if len(files) == num_files:
                    break

        return files

    def get_deletable_files(self, archive_name, delete_after, session=None):
        """
        Gets all files that are deletable, meaning that the local and remote
        md5sums match, and they have existed longer than ``delete_after``
        seconds.

        Args
        -----
            archive_name : str
                Name of archive to pull files from
            delete_after : float
                Time since creation (in seconds) for which it's ok to delete
                files.
            session : sqlalchemy session
                Session to use to query files.
        """
        if session is None:
            session = self.Session()

        query = session.query(SupRsyncFile).filter(
            SupRsyncFile.removed == None,
            SupRsyncFile.archive_name == archive_name,
            SupRsyncFile.deletable,
        )

        files = []
        now = time.time()
        for f in query.all():
            if f.local_md5sum == f.remote_md5sum:
                print(f.local_path, f.timestamp)
                if now > f.timestamp + delete_after:
                    files.append(f)

        return files


class SupRsyncFileHandler:
    """
    Helper class to handle files in the suprsync db and copy them to their
    dest / delete them if enough time has passed.
    """
    def __init__(self, file_manager, archive_name, remote_basedir,
                 ssh_host=None, ssh_key=None, cmd_timeout=None,
                 copy_timeout=None):
        self.srfm = file_manager
        self.archive_name = archive_name
        self.ssh_host = ssh_host
        self.ssh_key = ssh_key
        self.remote_basedir = remote_basedir
        self.log = txaio.make_logger()
        self.cmd_timeout = cmd_timeout
        self.copy_timeout = copy_timeout

    def run_on_remote(self, cmd, timeout=None):
        """
        Runs a command on the remote server or locally if ssh_host is None.

        Parameters
        -----------
        cmd : list
            Command to be run
        """
        _cmd = []
        if self.ssh_host is not None:
            _cmd += ['ssh', self.ssh_host]
            if self.ssh_key is not None:
                _cmd.extend(['-i', self.ssh_key])
        _cmd += cmd

        if timeout is None:
            timeout = self.cmd_timeout
        self.log.debug(f"Running: {' '.join(_cmd)}")
        res = subprocess.run(_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                             timeout=timeout, check=True)
        if res.stderr:
            self.log.error("stderr for cmd: {cmd}\n{err}",
                           cmd=_cmd, err=res.stderr.decode())
        return res

    def copy_files(self, max_copy_attempts=None, num_files=None):
        """
        Copies a batch of files, and computes remote md5sums.

        Args
        ----
            max_copy_attempts : int
                Max number of failed copy atempts
            num_files : int
                Number of files to return
        """
        with self.srfm.Session.begin() as session:
            files = self.srfm.get_copyable_files(
                self.archive_name, max_copy_attempts=max_copy_attempts,
                num_files=num_files, session=session
            )

            if not files:
                return

            if self.ssh_host is not None:
                dest = self.ssh_host + ':' + self.remote_basedir
            else:
                dest = self.remote_basedir

            # Creates temp directory with remote dir structure of symlinks for
            # rsync to copy.
            file_map = {}
            with tempfile.TemporaryDirectory() as tmp_dir:
                self.log.info("Copying files:")
                for file in files:
                    self.log.info(f"- {file.local_path}")
                    tmp_path = os.path.join(tmp_dir, file.remote_path)
                    os.makedirs(os.path.dirname(tmp_path), exist_ok=True)
                    os.symlink(file.local_path, tmp_path)

                    remote_path = os.path.normpath(
                        os.path.join(self.remote_basedir, file.remote_path)
                    )
                    file_map[remote_path] = file

                cmd = ['rsync', '-Lrt', tmp_dir+'/', dest]
                subprocess.run(cmd, check=True, timeout=self.copy_timeout)

            for file in files:
                file.copied = time.time()

            remote_paths = [
                os.path.join(self.remote_basedir, f.remote_path)
                for f in files
            ]

            res = self.run_on_remote(['md5sum'] + remote_paths)
            for line in res.stdout.decode().split('\n'):
                split = line.split()

                # If file cannot be found, line will say:
                # "md5sum: file: No such file or directory
                if len(split) != 2:
                    continue

                md5sum, path = line.split()
                file_map[os.path.normpath(path)].remote_md5sum = md5sum

    def delete_files(self, delete_after):
        """
        Gets deletable files, deletes them, and updates file info

        Args
        -----
            delete_after : float
                Time since creation (in seconds) for which it's ok to delete
                files.
        """
        with self.srfm.Session.begin() as session:
            files = self.srfm.get_deletable_files(
                self.archive_name, delete_after, session=session
            )
            for file in files:
                if os.path.exists(file.local_path):
                    try:
                        self.log.info(f"Removing file {file.local_path}")
                        os.remove(file.local_path)
                        file.removed = time.time()
                    except PermissionError:
                        self.log.error(
                            f"Permission error: Could not remove {file.local_path}"
                        )
                else:
                    self.log.warn(
                        "File {file.local_path} no longer exists! "
                        "Updating remove time to be 0"
                    )
                    file.removed = 0



# Lakeshore372.py

import sys
import socket
import time
import numpy as np

# Lookup keys for command parameters.
autorange_key = {'0': 'off',
                 '1': 'on',
                 '2': 'ROX 102B'}

mode_key = {'0': 'voltage',
            '1': 'current'}

mode_lock = {'voltage': '0',
             'current': '1'}

voltage_excitation_key = {1: 2.0e-6,
                          2: 6.32e-6,
                          3: 20.0e-6,
                          4: 63.2e-6,
                          5: 200.0e-6,
                          6: 632.0e-6,
                          7: 2.0e-3,
                          8: 6.32e-3,
                          9: 20.0e-3,
                          10: 63.2e-3,
                          11: 200.0e-3,
                          12: 632.0e-3}

current_excitation_key = {1: 1.0e-12,
                          2: 3.16e-12,
                          3: 10.0e-12,
                          4: 31.6e-12,
                          5: 100.0e-12,
                          6: 316.0e-12,
                          7: 1.0e-9,
                          8: 3.16e-9,
                          9: 10.0e-9,
                          10: 31.6e-9,
                          11: 100.0e-9,
                          12: 316.0e-9,
                          13: 1.0e-6,
                          14: 3.16e-6,
                          15: 10.0e-6,
                          16: 31.6e-6,
                          17: 100.0e-6,
                          18: 316.0e-6,
                          19: 1.0e-3,
                          20: 3.16e-3,
                          21: 10.0-3,
                          22: 31.6-3}

voltage_excitation_lock = {2.0e-6: 1,
                           6.32e-6: 2,
                           20.0e-6: 3,
                           63.2e-6: 4,
                           200.0e-6: 5,
                           632.0e-6: 6,
                           2.0e-3: 7,
                           6.32e-3: 8,
                           20.0e-3: 9,
                           63.2e-3: 10,
                           200.0e-3: 11,
                           632.0e-3: 12}

current_excitation_lock = {1.0e-12: 1,
                           3.16e-12: 2,
                           10.0e-12: 3,
                           31.6e-12: 4,
                           100.0e-12: 5,
                           316.0e-12: 6,
                           1.0e-9: 7,
                           3.16e-9: 8,
                           10.0e-9: 9,
                           31.6e-9: 10,
                           100.0e-9: 11,
                           316.0e-9: 12,
                           1.0e-6: 13,
                           3.16e-6: 14,
                           10.0e-6: 15,
                           31.6e-6: 16,
                           100.0e-6: 17,
                           316.0e-6: 18,
                           1.0e-3: 19,
                           3.16e-3: 20,
                           10.0-3: 21,
                           31.6-3: 22}

range_key = {1: 2.0e-3,
             2: 6.32e-3,
             3: 20.0e-3,
             4: 63.2e-3,
             5: 200e-3,
             6: 632e-3,
             7: 2.0,
             8: 6.32,
             9: 20.0,
             10: 63.2,
             11: 200,
             12: 632,
             13: 2e3,
             14: 6.32e3,
             15: 20.0e3,
             16: 63.2e3,
             17: 200e3,
             18: 632e3,
             19: 2e6,
             20: 6.32e6,
             21: 20.0e6,
             22: 63.2e6}

range_lock = {2.0e-3: 1,
              6.32e-3: 2,
              20.0e-3: 3,
              63.2e-3: 4,
              200e-3: 5,
              632e-3: 6,
              2.0: 7,
              6.32: 8,
              20.0: 9,
              63.2: 10,
              200: 11,
              632: 12,
              2e3: 13,
              6.32e3: 14,
              20.0e3: 15,
              63.2e3: 16,
              200e3: 17,
              632e3: 18,
              2e6: 19,
              6.32e6: 20,
              20.0e6: 21,
              63.2e6: 22}

units_key = {'1': 'kelvin',
             '2': 'ohms'}

units_lock = {'kelvin': '1',
              'ohms': '2'}

csshunt_key = {'0': 'on',
               '1': 'off'}

tempco_key = {'1': 'negative',
              '2': 'positive'}

tempco_lock = {'negative': '1',
               'positive': '2'}

format_key = {'3': "Ohm/K (linear)",
              '4': "log Ohm/K (linear)",
              '7': "Ohm/K (cubic spline)"}

format_lock = {"Ohm/K (linear)": '3',
               "log Ohm/K (linear)": '4',
               "Ohm/K (cubic spline)": '7'}

heater_range_key = {"0": "Off", "1": 31.6e-6, "2": 100e-6, "3": 316e-6,
                    "4": 1e-3, "5": 3.16e-3, "6": 10e-3, "7": 31.6e-3,
                    "8": 100e-3}
heater_range_lock = {v:k for k, v in heater_range_key.items()}
heater_range_lock["On"] = "1"

output_modes = {'0': 'Off', '1': 'Monitor Out', '2': 'Open Loop', '3': 'Zone', '4': 'Still',
                '5': 'Closed Loop', '6': 'Warm up'}
output_modes_lock = {v.lower():k for k, v in output_modes.items()}

heater_display_key = { '1': 'current',
                '2': 'power'}
heater_display_lock = {v: k for k,v in heater_display_key.items()}

def _establish_socket_connection(ip, timeout, port=7777):
    """Establish socket connection to the LS372.

    Parameters
    ----------
    ip : str
        IP address of the LS372
    timeout : int
        timeout period for the socket connection in seconds
    port : int
        Port for the connection, defaults to the default LS372 port of 7777

    Returns
    -------
    socket.socket
        The socket object with open connection to the 372

    """
    com = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        com.connect((ip, port))
    except OSError as e:
        if 'No route to host' in e.strerror:
            raise ConnectionError("Cannot connect to LS372")
        else:
            raise e
    com.settimeout(timeout)

    return com


class LS372:
    """
        Lakeshore 372 class.

    Attributes:
        channels - list of channels, index corresponds to channel number with
                   index 0 corresponding to the control channel, 'A'
    """
    def __init__(self, ip, timeout=10, num_channels=16):
        self.com = _establish_socket_connection(ip, timeout)
        self.num_channels = num_channels

        self.id = self.get_id()
        self.autoscan = self.get_autoscan()
        # Enable all channels
        #  going to hold off on enabling all channels automatically - bjk
        # for i in range(self.num_channels):
        #     print(i)
        #     self.msg('INSET %d,1,3,3'%(i))

        self.channels = []
        for i in range(num_channels + 1):
            if i == 0:
                c = Channel(self, 'A')
            else:
                c = Channel(self, i)
            self.channels.append(c)

        self.sample_heater = Heater(self, 0)
        self.still_heater = Heater(self, 2)

    def msg(self, message):
        """Send message to the Lakeshore 372 over ethernet.

        If we're asking for something from the Lakeshore (indicated by a ? in
        the message string), then we will attempt to ask twice before giving up
        due to potential communication timeouts.

        Parameters
        ----------
        message : str
            Message string as described in the Lakeshore 372 manual.

        Returns
        -------
        str
            Response string from the Lakeshore, if any. Else, an empty string.

        """
        msg_str = f'{message}\r\n'.encode()

        if '?' in message:
            self.com.send(msg_str)
            # Try once, if we timeout, try again. Usually gets around single event glitches.
            for attempt in range(2):
                try:
                    resp = str(self.com.recv(4096), 'utf-8').strip()
                    break
                except socket.timeout:
                    print("Warning: Caught timeout waiting for response to '%s', trying again " \
                          "before giving up"%message)
                    if attempt == 1:
                        raise RuntimeError('Query response to Lakeshore timed out after two ' \
                                           'attempts. Check connection.')
        else:
            self.com.send(msg_str)
            resp = ''

        return resp

    def get_id(self):
        """Get the ID number of the Lakeshore unit."""
        return self.msg('*IDN?')

    def get_temp(self, unit=None, chan=-1):
        """Get temperature from the Lakeshore.

        Args:
            unit (str): Unit to return reading for ('ohms' or 'kelvin')
            chan (str): Channel to query, -1 for currently active channel

        Returns:
            float: The reading from the lakeshore, either in ohms or kelvin.

        """
        if (chan == -1):
            resp = self.msg("SCAN?")
            c = resp.split(',')[0]
        elif (chan == 0):
            c = 'A'
        else:
            c = str(chan)

        if unit is None:
            active_ch = self.get_active_channel()
            unit = active_ch.units

        assert unit.lower() in ['ohms', 'kelvin']

        if unit == 'ohms':
            # Sensor is same as Resistance Query
            return float(self.msg('SRDG? %s' % c))
        if unit == 'kelvin':
            return float(self.msg('KRDG? %s' % c))

    def get_autoscan(self):
        """Determine state of autoscan.

        :returns: state of autoscanner
        :rtype: bool
        """
        resp = self.msg('SCAN?')
        scan_state = bool(int(resp.split(',')[1]))
        self.autoscan = scan_state
        return scan_state

    def _set_autoscan(self, start=1, autoscan=0):
        """Set the autoscan state and start channel for scanning.

        :param start: Channel number to start scanning
        :type start: int
        :param autoscan: State of autoscan, 0 for off, 1 for on
        :type autoscan: int
        """
        assert autoscan in [0, 1]

        self.msg('SCAN {},{}'.format(start, autoscan))
        self.autoscan = bool(autoscan)

    def enable_autoscan(self):
        """Enable the autoscan feature of the Lakeshore 372.

        Will query active channel to pass already selected channel to SCAN
        command.
        """
        active_channel = self.get_active_channel()
        self.msg('SCAN {},{}'.format(active_channel.channel_num, 1))
        self.autoscan = True

    def disable_autoscan(self):
        """Disable the autoscan feature of the Lakeshore 372.

        Will query active channel to pass already selected channel to SCAN
        command.
        """
        active_channel = self.get_active_channel()
        self.msg('SCAN {},{}'.format(active_channel.channel_num, 0))
        self.autoscan = False

    def get_active_channel(self):
        """Query the Lakeshore for which channel it's currently scanning.

        :returns: channel object describing the scanned channel
        :rtype: Channel Object
        """
        resp = self.msg("SCAN?")
        channel_number = int(resp.split(',')[0])
        channel_list = [_.channel_num for _ in self.channels]
        idx = channel_list.index(channel_number)
        return self.channels[idx]

    def set_active_channel(self, channel):
        """Set the active scanner channel.

        Query using SCAN? to determine autoscan parameter and set active
        channel.

        :param channel: Channel number to switch scanner to. 1-8 or 1-16
                        depending on scanner type
        :type channel: int
        """
        resp = self.msg("SCAN?")
        autoscan_setting = resp.split(',')[1]
        self.msg('SCAN {},{}'.format(channel, autoscan_setting))

    # NET?
    def get_network_settings(self):
        pass

    # NETID?
    def get_network_configuration(self):
        pass


class Channel:
    """Lakeshore 372 Channel Object

    :param ls: Lakeshore unit for communication
    :type ls: LS372 Object
    :param channel_num: The channel number (1-8 or 1-16 depending on scanner
                        type)
    :type channel_num: int
    """
    def __init__(self, ls, channel_num):
        self.ls = ls
        self.channel_num = channel_num
        self.enabled = False
        self.get_input_channel_parameter()
        self.get_sensor_input_name()
        self.get_input_setup()
        self.tlimit = self.get_temperature_limit()

    def get_input_channel_parameter(self):
        """Run Input Channel Parameter Query

        ::

          Input channel parameters include:
              off/on - Specifies whether the input/channel is disabled or enabled
                  type off/on - bool
              dwell - Specifies a value for the autoscanning dwell time 1 to 200 s
                  type dwell - int in units of seconds
              pause - Specifies a value for the change pause time: 3 to 200 s
                  type pause - int in units of seconds
              curve number - Specifies which curve the channel uses
                  type curve number - int
              tempco - Sets the temperature coefficien that will be used for
                       temperature control if no curve is selected
                  type tempco - str


        :returns: response from INSET? command

        Reference: LakeShore 372 Manual - pg177
        """
        resp = self.ls.msg(f"INSET? {self.channel_num}").split(',')

        self.enabled = bool(int(resp[0]))
        self.dwell = int(resp[1])  # seconds
        self.pause = int(resp[2])  # seconds
        self.curve_num = int(resp[3])
        self.tempco = tempco_key[resp[4]]

        return resp

    def _set_input_channel_parameter(self, params):
        """Set INSET.

        Parameters should be <disabled/enabled>, <dwell>, <pause>, <curve
        number>, <tempco>. Will determine <input/channel> from attributes. This
        allows us to use output from get_input_channel_parameters directly, as
        it doesn't return <input/channel>.

        :param params: INSET parameters
        :type params: list of str

        :returns: response from ls.msg
        """
        assert len(params) == 5

        reply = [str(self.channel_num)]
        [reply.append(x) for x in params]

        param_str = ','.join(reply)
        return self.ls.msg(f"INSET {param_str}")

    def get_input_setup(self):
        """Run Input Setup Query, storing results in human readable format.

        ::

          Input setup parameters include:
              mode - Sensor excitation mode.
                     Measurement input: 0 = Voltage Excitation Mode,
                                        1 = Current Excitation Mode
                     Control input (channel A): 1 = Current Excitation
                  type mode - int
              excitation - Measurement input excitation range
                  type excitation - int
              autorange - Specifies if auto range is enabled.
                              0 = off,
                              1 = autorange current,
                              2 = ROX102B Autorange (control input only)
                  type autorange - int
              range - Measurement input resistance. Ignored for control input.
                  type range - int
              cs shunt - Current source shunt.
                          0 = current source not shunted, excitation on
                          1 = current source shunted, excitation off
                  type cs shunt - int
              units - Specifies the preferred units parameter for sensor readings
                      and for the control setpoint:
                          1 = kelvin,
                          2 = ohms
                  type units - int

        :returns: response from INTYPE? command

        Reference: LakeShore 372 Manual - pg178-179
        """
        resp = self.ls.msg(f"INTYPE? {self.channel_num}").split(',')

        _mode = resp[0]
        _excitation = resp[1]
        _autorange = resp[2]
        _range = resp[3]
        _csshunt = resp[4]
        _units = resp[5]

        if self.channel_num == "A":
            control_channel = True
        else:
            control_channel = False

        self.mode = mode_key[_mode]

        if control_channel:
            # TODO: should test this with control channel connected
            excitation_key = {'1': {1: 316.0e-12,
                                    2: 1e-9,
                                    3: 3.16e-9,
                                    4: 10e-9,
                                    5: 31.6e-9,
                                    6: 100e-9}}
        else:
            excitation_key = {'0': voltage_excitation_key,
                              '1': current_excitation_key}

        excitation_units_key = {'0': 'volts',
                                '1': 'amps'}

        self.excitation = excitation_key[_mode][int(_excitation)]
        self.excitation_units = excitation_units_key[_mode]

        self.autorange = autorange_key[_autorange]

        self.range = range_key[int(_range)]

        self.csshunt = csshunt_key[_csshunt]

        self.units = units_key[_units]

        return resp

    def _set_input_setup(self, params):
        """Set INTYPE.

        Parameters are <mode>, <excitation>, <autorange>, <range>, <cs shunt>,
        <units>. Will determine <input/channel> from attributes.

        :param params: INTYPE parameters
        :type params: list of str

        :returns: response from ls.msg
        """
        assert len(params) == 6

        reply = [str(self.channel_num)]
        [reply.append(x) for x in params]

        param_str = ','.join(reply)
        return self.ls.msg(f"INTYPE {param_str}")

    # Public API

    def get_excitation_mode(self):
        """Get the excitation mode form INTYPE?

        :returns: excitation mode, 'current' or 'voltage'
        :rtype: str
        """
        resp = self.get_input_setup()
        self.mode = mode_key[resp[0]]
        return self.mode

    def set_excitation_mode(self, excitation_mode):
        """Set the excitation mode to either voltage excitation or current
        exitation.

        :param excitation_mode: mode we want, must be 'current' or 'voltage'
        :type excitation_mode: str

        :returns: reply from INTYPE call
        :rtype: str

        """
        assert excitation_mode in ['voltage', 'current']

        resp = self.get_input_setup()
        resp[0] = mode_lock[excitation_mode]

        self.mode = mode_key[resp[0]]

        return self._set_input_setup(resp)

    def get_excitation(self):
        """Get excitation value from INTYPE?

        :returns: excitation value in volts or amps, depending on mode
        :rtype: float
        """
        resp = self.get_input_setup()
        _mode = resp[0]
        _excitation = resp[1]

        if self.channel_num == "A":
            control_channel = True
        else:
            control_channel = False

        if control_channel:
            excitation_key = {'1': {1: 316.0e-12,
                                    2: 1e-9,
                                    3: 3.16e-9,
                                    4: 10e-9,
                                    5: 31.6e-9,
                                    6: 100e-9}}
        else:
            excitation_key = {'0': voltage_excitation_key,
                              '1': current_excitation_key}

        self.excitation = excitation_key[_mode][int(_excitation)]

        return self.excitation

    def set_excitation(self, excitation_value):
        """Set voltage/current exitation to specified value via INTYPE command.

        :param excitation_value: value in volts/amps of excitation
        :type excitation_value: float

        :returns: response from INTYPE command
        :rtype: str
        """
        _mode = self.mode

        if self.channel_num == "A":
            control_channel = True
        else:
            control_channel = False

        if control_channel:
            excitation_lock = {316.0e-12: 1,
                               1e-9: 2,
                               3.16e-9: 3,
                               10e-9: 4,
                               31.6e-9: 5,
                               100e-9: 6}
        else:
            if _mode == 'voltage':
                excitation_lock = voltage_excitation_lock
            elif _mode == 'current':
                excitation_lock = current_excitation_lock

        closest_value = min(excitation_lock, key=lambda x: abs(x-excitation_value))

        resp = self.get_input_setup()
        resp[1] = str(excitation_lock[closest_value])

        return self._set_input_setup(resp)

    def enable_autorange(self):
        """Enable auto range for channel via INTYPE command."""
        resp = self.get_input_setup()
        resp[2] = '1'
        self.autorange = autorange_key[resp[2]]
        return self._set_input_setup(resp)

    def disable_autorange(self):
        """Disable auto range for channel via INTYPE command."""
        resp = self.get_input_setup()
        resp[2] = '0'
        self.autorange = autorange_key[resp[2]]
        return self._set_input_setup(resp)

    def set_resistance_range(self, resistance_range):
        """Set the resistance range.

        :param resistance_range: range in ohms we want to measure. Doesn't need
                                 to be exactly one of the options on the
                                 lakeshore, will select closest valid range,
                                 though note these are in increments of 2, 6.32, 20, 63.2, etc.
        :type resistance_range: float

        :returns: response from INTYPE command
        :rtype: str
        """

        def get_closest_resistance_range(num):
            """Gets the closest valid resistance range."""
            ranges = [2.0e-3, 6.32e-3, 20.0e-3, 63.2e-3, 200e-3, 632e-3, 2.0,
                      6.32, 20.0, 63.2, 200, 632, 2e3, 6.32e3, 20.0e3, 63.2e3,
                      200e3, 632e3, 2e6, 6.32e6, 20.0e6, 63.2e6]

            return min(ranges, key=lambda x: abs(x-num))

        _range = get_closest_resistance_range(resistance_range)

        resp = self.get_input_setup()
        resp[3] = str(range_lock[_range])
        self.range = _range
        return self._set_input_setup(resp)

    def get_resistance_range(self):
        """Get the resistance range.

        :returns: resistance range in Ohms
        :rtype: float
        """
        resp = self.get_input_setup()
        _range = resp[3]
        self.range = range_key[int(_range)]
        return self.range

    def enable_excitation(self):
        """Enable excitation by not shunting the current source via INTYPE command.

        :returns: state of excitation
        :rtype: str
        """
        resp = self.get_input_setup()
        resp[4] = '0'
        self.csshunt = csshunt_key[resp[4]]
        return self._set_input_setup(resp)

    def disable_excitation(self):
        """Disable excitation by shunting the current source via INTYPE command.

        :returns: state of excitation
        :rtype: str
        """
        resp = self.get_input_setup()
        resp[4] = '1'
        self.csshunt = csshunt_key[resp[4]]
        return self._set_input_setup(resp)

    def get_excitation_power(self):
        """Get the most recent power calculation for the channel via RDGPWR? command.

        :returns: power in Watts
        :rtype: float
        """
        # TODO: Confirm units on this are watts
        resp = self.ls.msg(f"RDGPWR? {self.channel_num}").strip()
        return float(resp)

    def set_units(self, units):
        """Set preferred units using INTYPE command.

        :param units: preferred units parameter for sensor readings, 'kelvin'
                      or 'ohms'
        :type units: str

        :returns: response from INTYPE command
        :rtype: str
        """
        assert units.lower() in ['kelvin', 'ohms']

        resp = self.get_input_setup()
        resp[5] = units_lock[units.lower()]
        return self._set_input_setup(resp)

    def get_units(self):
        """Get preferred units from INTYPE? command.

        :returns: preferred units
        :rtype: str
        """
        resp = self.get_input_setup()
        _units = resp[5]
        self.units = units_key[_units]

        return self.units

    def enable_channel(self):
        """Enable channel using INSET command.

        :returns: response from self._set_input_channel_parameter()
        :rtype: str
        """
        resp = self.get_input_channel_parameter()
        resp[0] = '1'
        self.enabled = True
        return self._set_input_channel_parameter(resp)

    def disable_channel(self):
        """Disable channel using INSET command.

        :returns: response from self._set_input_channel_parameter()
        :rtype: str
        """
        resp = self.get_input_channel_parameter()
        resp[0] = '0'
        self.enabled = False
        return self._set_input_channel_parameter(resp)

    def set_dwell(self, dwell):
        """Set the autoscanning dwell time.

        :param dwell: Dwell time in seconds
        :type dwell: int

        :returns: response from self._set_input_channel_parameter()
        :rtype: str
        """
        assert dwell in range(1, 201), "Dwell must be 1 to 200 sec"

        resp = self.get_input_channel_parameter()
        resp[1] = str(dwell)  # seconds
        self.dwell = dwell  # seconds
        return self._set_input_channel_parameter(resp)

    def get_dwell(self):
        """Get the autoscanning dwell time.

        :returns: the dwell time in seconds
        :rtype: int
        """
        resp = self.get_input_channel_parameter()
        self.dwell = int(resp[1])
        return self.dwell

    def set_pause(self, pause):
        """Set pause time.

        :param pause: Pause time in seconds
        :type pause: int

        :returns: response from self._set_input_channel_parameter()
        :rtype: str
        """
        assert pause in range(3, 201), "Pause must be 3 to 200 sec"

        resp = self.get_input_channel_parameter()
        resp[2] = str(pause)  # seconds
        self.pause = pause  # seconds
        return self._set_input_channel_parameter(resp)

    def get_pause(self):
        """Get the pause time from INSET.

        :returns: the pause time in seconds
        :rtype: int
        """
        resp = self.get_input_channel_parameter()
        self.pause = int(resp[2])  # seconds
        return self.pause

    def set_calibration_curve(self, curve_number):
        """Set calibration curve using INSET.

        Note: If curve doesn't exist, curve number gets set to 0.

        :param curve_number: Curve number for temperature conversion
        :type curve_number: int
        """
        assert curve_number in range(0, 60), "Curve number must from 0 to 59"

        resp = self.get_input_channel_parameter()
        resp[3] = str(curve_number)
        self.curve_num = self.get_calibration_curve()
        return self._set_input_channel_parameter(resp)

    def get_calibration_curve(self):
        """Get calibration curve number using INSET?

        :returns: curve number in use for the channel
        :rtype: int
        """
        resp = self.get_input_channel_parameter()
        self.curve_num = int(resp[3])
        return self.curve_num

    def set_temperature_coefficient(self, coefficient):
        """Set tempertaure coefficient with INSET.

        :param coefficient: set coefficient to be used for temperature control
                            if no curve is selected, either 'negative' or
                            'positive'
        :type coefficient: str

        :returns: response from _set_input_channel_parameter()
        :rtype: str
        """
        assert coefficient in ['positive', 'negative']

        resp = self.get_input_channel_parameter()
        resp[4] = tempco_lock[coefficient]
        self.tempco = coefficient
        return self._set_input_channel_parameter(resp)

    def get_temperature_coefficient(self):
        """Get temperature coefficient from INSET?

        :returns: temperature coefficient
        """
        resp = self.get_input_channel_parameter()
        self.tempco = tempco_key[resp[4]]
        return self.tempco

    def get_sensor_input_name(self):
        """Run Sensor Input Name Query

        :returns: response from INNAME? command
        :rtype: str
        """
        resp = self.ls.msg(f"INNAME? {self.channel_num}").strip()

        self.name = resp

        return resp

    def set_sensor_input_name(self, name):
        """Set sensor input name using INNAME.

        Note: ',' and ';' characters are sanatized from input

        :param name: name to give input channel
        :type name: str
        """
        name = name.replace(',', '').replace(';', '')
        resp = self.ls.msg(f'INNAME {self.channel_num},"{name}"')
        self.name = name
        return resp

    def get_kelvin_reading(self):
        """Get temperature reading from channel.

        :returns: temperature from channel in Kelvin
        :rtype: float
        """
        return float(self.ls.msg(f"RDGK? {self.channel_num}"))

    def get_resistance_reading(self):
        """Get resistence reading from channel.

        :returns: resistance from channel in Ohms
        :rtype: float
        """
        return float(self.ls.msg(f"RDGR? {self.channel_num}"))

    def get_reading_status(self):
        """Get status of input reading.

        :returns: list of errors on reading (or None if no errors)
        :rtype: list of str
        """
        resp = self.ls.msg(f"RDGST? {self.channel_num}")
        error_sum = int(resp)

        errors = {128: "T.UNDER",
                  64: "T.OVER",
                  32: "R.UNDER",
                  16: "R.OVER",
                  8: "VDIF OVL",
                  4: "VMIX OVL",
                  2: "VCM OVL",
                  1: "CS OVL"}

        error_list = []
        for key, value in errors.items():
            if key <= error_sum:
                error_list.append(value)
                error_sum -= key

        assert error_sum == 0

        if len(error_list) == 0:
            error_list = None

        return error_list

    def get_sensor_reading(self):
        """Get sensor reading from channel.

        :returns: resistance from channel in Ohms
        :rtype: float
        """
        return float(self.ls.msg(f"SRDG? {self.channel_num}"))

    def set_temperature_limit(self, limit):
        """Set temperature limit in kelvin for which to shutdown all control
        outputs when exceeded. A temperature limit of zero turns the
        temperature limit feature off for the given sensor input.

        :param limit: temperature limit in kelvin
        :type limit: float

        :returns: response from TLIMIT command
        :rtype: str
        """
        resp = self.ls.msg(f"TLIMIT {self.channel_num},{limit}")
        self.tlimit = limit
        return resp

    def get_temperature_limit(self):
        """Get temperature limit, at which output controls are shutdown.

        A temperature limit of 0 disables this feature.

        :returns: temperature limit in Kelvin
        :rtype: float
        """
        resp = self.ls.msg(f"TLIMIT? {self.channel_num}").strip()
        self.tlimit = float(resp)
        return self.tlimit

    def __str__(self):
        string = "-" * 50 + "\n"
        string += "Channel %s: %s\n" % (self.channel_num, self.name)
        string += "-" * 50 + "\n"
        string += "\t%-30s\t%r\n" % ("Enabled :", self.enabled)
        string += "\t%-30s\t%s %s\n" % ("Dwell:", self.dwell, "seconds")
        string += "\t%-30s\t%s %s\n" % ("Pause:", self.pause, "seconds")
        string += "\t%-30s\t%s\n" % ("Curve Number:", self.curve_num)
        string += "\t%-30s\t%s\n" % ("Temperature Coefficient:", self.tempco)
        string += "\t%-30s\t%s\n" % ("Excitation State:", self.csshunt)
        string += "\t%-30s\t%s\n" % ("Excitation Mode:", self.mode)
        string += "\t%-30s\t%s %s\n" % ("Excitation:", self.excitation, self.excitation_units)
        string += "\t%-30s\t%s\n" % ("Autorange:", self.autorange)
        string += "\t%-30s\t%s %s\n" % ("Resistance Range:", self.range, "ohms")
        string += "\t%-30s\t%s\n" % ("Preferred Units:", self.units)

        return string


class Curve:
    """Calibration Curve class for the LS372."""
    def __init__(self, ls, curve_num):
        self.ls = ls
        self.curve_num = curve_num

        self.name = None
        self.serial_number = None
        self.format = None
        self.limit = None
        self.coefficient = None
        self.get_header()  # populates above values

    def get_header(self):
        """Get curve header description.

        :returns: response from CRVHDR? in list
        :rtype: list of str
        """
        resp = self.ls.msg(f"CRVHDR? {self.curve_num}").split(',')

        _name = resp[0].strip()
        _sn = resp[1].strip()
        _format = resp[2]
        _limit = float(resp[3])
        _coefficient = resp[4]

        self.name = _name
        self.serial_number = _sn

        self.format = format_key[_format]

        self.limit = _limit
        self.coefficient = tempco_key[_coefficient]

        return resp

    def _set_header(self, params):
        """Set the Curve Header with the CRVHDR command.

        Parameters should be <name>, <SN>, <format>, <limit value>,
        <coefficient>. We will determine <curve> from attributes. This
        allows us to use output from get_header directly, as it doesn't return
        the curve number.

        <name> is limited to 15 characters. Longer names take the fist 15 characters
        <sn> is limited to 10 characters. Longer sn's take the last 10 digits

        :param params: CRVHDR parameters
        :type params: list of str

        :returns: response from ls.msg
        """
        assert len(params) == 5

        _curve_num = self.curve_num
        _name = params[0][:15]
        _sn = params[1][-10:]
        _format = params[2]
        assert _format.strip() in ['3', '4', '7']
        _limit = params[3]
        _coeff = params[4]
        assert _coeff.strip() in ['1', '2']

        return self.ls.msg(f'CRVHDR {_curve_num},"{_name}","{_sn}",{_format},{_limit},{_coeff}')

    def get_name(self):
        """Get the curve name with the CRVHDR? command.

        :returns: The curve name
        :rtype: str
        """
        self.get_header()
        return self.name

    def set_name(self, name):
        """Set the curve name with the CRVHDR command.

        :param name: The curve name, limit of 15 characters, longer names get truncated
        :type name: str

        :returns: the response from the CRVHDR command
        :rtype: str
        """
        resp = self.get_header()
        resp[0] = name.upper()
        self.name = resp[0]
        return self._set_header(resp)

    def get_serial_number(self):
        """Get the curve serial number with the CRVHDR? command."

        :returns: The curve serial number
        :rtype: str
        """
        self.get_header()
        return self.serial_number

    def set_serial_number(self, serial_number):
        """Set the curve serial number with the CRVHDR command.

        :param serial_number: The curve serial number, limit of 10 characters,
                              longer serials get truncated
        :type name: str

        :returns: the response from the CRVHDR command
        :rtype: str
        """
        resp = self.get_header()
        resp[1] = serial_number
        self.serial_number = resp[1]
        return self._set_header(resp)

    def get_format(self):
        """Get the curve data format with the CRVHDR? command."

        :returns: The curve data format
        :rtype: str
        """
        self.get_header()
        return self.format

    def set_format(self, _format):
        """Set the curve format with the CRVHDR command.

        :param _format: The curve format, valid formats are:
                          "Ohm/K (linear)"
                          "log Ohm/K (linear)"
                          "Ohm/K (cubic spline)"
        :type name: str

        :returns: the response from the CRVHDR command
        :rtype: str
        """
        resp = self.get_header()

        assert _format in format_lock.keys(), "Please select a valid format"

        resp[2] = format_lock[_format]
        self.format = _format
        return self._set_header(resp)

    def get_limit(self):
        """Get the curve temperature limit with the CRVHDR? command.

        :returns: The curve temperature limit
        :rtype: str
        """
        self.get_header()
        return self.limit

    def set_limit(self, limit):
        """Set the curve temperature limit with the CRVHDR command.

        :param limit: The curve temperature limit
        :type limit: float

        :returns: the response from the CRVHDR command
        :rtype: str
        """
        resp = self.get_header()
        resp[3] = str(limit)
        self.limit = limit
        return self._set_header(resp)

    def get_coefficient(self):
        """Get the curve temperature coefficient with the CRVHDR? command.

        :returns: The curve temperature coefficient
        :rtype: str
        """
        self.get_header()
        return self.coefficient

    def set_coefficient(self, coefficient):
        """Set the curve temperature coefficient with the CRVHDR command.

        :param coefficient: The curve temperature coefficient, either 'positive' or 'negative'
        :type limit: str

        :returns: the response from the CRVHDR command
        :rtype: str
        """
        assert coefficient in ['positive', 'negative']

        resp = self.get_header()
        resp[4] = tempco_lock[coefficient]
        self.tempco = coefficient
        return self._set_header(resp)

    def get_data_point(self, index):
        """Get a single data point from a curve, given the index, using the
        CRVPT? command.

        The format for the return value, a 3-tuple of floats, is chosen to work
        with how the get_curve() method later stores the entire curve in a
        numpy structured array.

        :param index: index of breakpoint to query
        :type index: int

        :returns: (units, tempertaure, curvature) values for the given breakpoint
        :rtype: 3-tuple of floats
        """
        resp = self.ls.msg(f"CRVPT? {self.curve_num},{index}").split(',')
        _units = float(resp[0])
        _temp = float(resp[1])
        return (_units, _temp)

    def _set_data_point(self, index, units, kelvin, curvature=None):
        """Set a single data point with the CRVPT command.

        :param index: data point index
        :type index: int
        :param units: value of the sensor units to 6 digits
        :type units: float
        :param kelvin: value of the corresponding temp in Kelvin to 6 digits
        :type kelvin: float
        :param curavature: used for calculating cublic spline coefficients (optional)
        :type curvature: float

        :returns: response from the CRVPT command
        :rtype: str
        """
        if curvature is None:
            resp = self.ls.msg(f"CRVPT {self.curve_num}, {index}, {units}, {kelvin}")
        else:
            resp = self.ls.msg(f"CRVPT {self.curve_num}, {index}, {units}, {kelvin}, {curvature}")

        return resp

    # Public API Elements
    def get_curve(self, _file=None):
        """Get a calibration curve from the LS372.

        If _file is not None, save to file location.
        """
        breakpoints = []
        for i in range(1, 201):
            x = self.get_data_point(i)
            if x[0] == 0:
                break
            breakpoints.append(x)

        struct_array = np.array(breakpoints, dtype=[('units', 'f8'),
                                                    ('temperature', 'f8')])

        self.breakpoints = struct_array

        if _file is not None:
            with open(_file, 'w') as f:
                f.write('Sensor Model:\t' + self.name + '\r\n')
                f.write('Serial Number:\t' + self.serial_number + '\r\n')
                f.write('Data Format:\t' + format_lock[self.format] + f'\t({self.format})\r\n')
                f.write('SetPoint Limit:\t%s\t(Kelvin)\r\n' % '%0.4f' % np.max(self.breakpoints['temperature']))
                f.write('Temperature coefficient:\t' + tempco_lock[self.coefficient] + f' ({self.coefficient})\r\n')
                f.write('Number of Breakpoints:\t%s\r\n' % len(self.breakpoints))
                f.write('\r\n')
                f.write('No.\tUnits\tTemperature (K)\r\n')
                f.write('\r\n')
                for idx, point in enumerate(self.breakpoints):
                    f.write('%s\t%s %s\r\n' % (idx+1, '%0.4f' % point['units'], '%0.4f' % point['temperature']))

        return self.breakpoints

    def set_curve(self, _file):
        """Set a calibration curve, loading it from the file.

        :param _file: the file to load the calibration curve from
        :type _file: str

        :returns: return the new curve header, refreshing the attributes
        :rtype: list of str
        """
        with open(_file) as f:
            content = f.readlines()

        header = []
        for i in range(0, 6):
            if i < 2 or i > 4:
                header.append(content[i].strip().split(":", 1)[1].strip())
            else:
                header.append(content[i].strip().split(":", 1)[1].strip().split("(", 1)[0].strip())

        # Skip to the R and T values in the file and strip them of tabs, newlines, etc
        values = []
        for i in range(9, len(content)):
            values.append(content[i].strip().split())

        self.delete_curve()  # remove old curve first, so old breakpoints don't remain

        self._set_header(header[:-1])  # ignore num of breakpoints

        for point in values:
            print("uploading %s"%point)
            self._set_data_point(point[0], point[1], point[2])

        # refresh curve attributes
        self.get_header()
        self._check_curve(_file)

    def _check_curve(self, _file):
        """After setting a data point for calibration curve,
        use CRVPT? command from get_data_point() to check
        that all points of calibration curve  were uploaded.
        If not, re-upload points.

        :param _file: calibration curve file
        :type _file: str
        """

        with open(_file) as f:
            content = f.readlines()

        #skipping header info
        values = []
        for i in range(9, len(content)):
            values.append(content[i].strip().split()) #data points that should have been uploaded
    
        for j in range(1, len(values)+1):
            try:
                resp = self.get_data_point(j) #response from the 372
                point = values[j-1]
                units = float(resp[0])
                temperature = float(resp[1])
                assert units == float(point[1]), "Point number %s not uploaded"%point[0]
                assert temperature == float(point[2]), "Point number %s not uploaded"%point[0]
                print("Successfully uploaded %s, %s" %(units,temperature))
            #if AssertionError, tell 372 to re-upload points
            except AssertionError:
                if units != float(point[1]):
                    self.set_curve(_file)

    def delete_curve(self):
        """Delete the curve using the CRVDEL command.

        :returns: the response from the CRVDEL command
        :rtype: str
        """
        resp = self.ls.msg(f"CRVDEL {self.curve_num}")
        self.get_header()
        return resp

    def __str__(self):
        string = "-" * 50 + "\n"
        string += "Curve %d: %s\n" % (self.curve_num, self.name)
        string += "-" * 50 + "\n"
        string += "  %-30s\t%r\n" % ("Serial Number:", self.serial_number)
        string += "  %-30s\t%s (%s)\n" % ("Format :", format_lock[self.format], self.format)
        string += "  %-30s\t%s\n" % ("Temperature Limit:", self.limit)
        string += "  %-30s\t%s\n" % ("Temperature Coefficient:", self.coefficient)

        return string


class Heater:
    """Heater class for LS372 control

    :param ls: the lakeshore object we're controlling
    :type ls: Lakeshore372.LS372
    :param output: the heater output we want to control, 0 = sample,
                   1 = warm-up, 2 = still
    :type output: int
    """
    def __init__(self, ls, output):
        self.ls = ls
        self.output = output
        self.mode = None
        self.input = None
        self.powerup = None
        self.polarity = None
        self.filter = None
        self.delay = None

        self.range = None

        self.resistance = None
        self.max_current = None
        self.max_user_current = None
        self.display = None

        self.get_output_mode()
        self.get_heater_setup()

    def get_output_mode(self):
        """Query the heater mode using the OUTMODE? command.

        :returns: 6-tuple with output mode, input channel, whether powerup is
                  enabled, polarity, unfiltered/filtered, and the autoscanning
                  delay time.
        :rtype: tuple
        """
        resp = self.ls.msg(f"OUTMODE? {self.output}").split(",")

        # TODO: make these human readable
        self.mode = output_modes[resp[0]]
        self.input = resp[1]

        self.powerup = resp[2]
        self.polarity = resp[3]
        self.filter = resp[4]
        self.delay = resp[5]

        return resp

    # OUTMODE
    def _set_output_mode(self, params):
        """Set the output mode of the heater with the OUTMODE command.

        Parameters should be <mode>, <input/channel>, <powerup enable>, <polarity>,
        <filter>, <delay>. Will determine <output> from attributes. This allows
        us to use output from get_output_mode directly, as it doesn't return
        <outpu>.

        :param params: OUTMODE parameters
        :type params: list of str

        :returns: response from ls.msg
        """
        assert len(params) == 6

        reply = [str(self.output)]
        [reply.append(x) for x in params]

        param_str = ','.join(reply)
        return self.ls.msg(f"OUTMODE {param_str}")

    def _set_heater_setup(self, params):
        """
        Sets the heater setup using the HTRSET command.

        Params must be a list with the parameters:
            <heater resistance>:    Heater load in Ohms (Sample);
                                    1=25 Ohms, 2=50 Ohms (warmp-up)
            <max current>: Specifies max heater output for warm-up heater.
                           0=User spec, 1=0.45 A, 2=0.63 A.
            <max user current>: Max heater output if max_current is set to user (warm-up only)
            <current/power>:    Specifies if heater display is current or power.
                                1=current, 2=power.

        :param params:
        :return:
        """
        assert self.output in [0,1]

        assert len(params) == 4

        reply = [str(self.output)]
        [reply.append(x) for x in params]

        param_str = ','.join(reply)
        return self.ls.msg("HTRSET {}".format(param_str))

    def get_mode(self):
        """Set output mode with OUTMODE? commnd.

        :returns: The output mode
        :rtype: str
        """
        self.get_output_mode()
        return self.mode

    def set_mode(self, mode):
        """Set output mode with OUTMODE commnd.

        :param mode: control mode for heater, see page 171 of LS372 Manual for
                     valid modes, as it changes per output
        :type mode: str

        :returns: the response from the OUTMODE command
        """
        # TODO: Make assertions check specific output and it's validity in mode selection
        assert mode.lower() in output_modes_lock.keys(), f"{mode} not a valid mode"

        resp = self.get_output_mode()
        resp[0] = output_modes_lock[mode.lower()]
        self.mode = mode
        return self._set_output_mode(resp)

    def get_manual_out(self):
        resp = self.ls.msg("MOUT? {}".format(self.output))
        return float(resp)


    def get_input_channel(self):
        """Get the control channel with the OUTMODE? command.

        :returns: The control channel
        :rtype: str
        """
        self.get_output_mode()
        return self.input

    def set_input_channel(self, _input):
        """Set the control channel with the OUTMODE command.

        :param _input: specifies which input or channel to control from
        :type _input: str or int
        """
        assert int(_input) in range(17) or _input == "A", f"{_input} not a valid input/channel"

        resp = self.get_output_mode()
        resp[1] = str(_input)
        self.input = str(_input)
        return self._set_output_mode(resp)

    def get_powerup(self):
        pass

    def set_powerup(self, powerup):
        """
        :param powerup: specifies whether the output remains on or shuts off
                        after power cycle. True for on after powerup
        :type powerup: bool
        """
        # assert powerup in [True, False], f"{powerup} not valid powerup parameter"
        # set_powerup = str(int(powerup))
        #
        pass

    def get_polarity(self):
        pass

    def set_polarity(self):
        """
        :param polarity: specifies output polarity: 'unipolar' or 'bipolar'
        :type polarity: str
        """
        # polarity_key = {0: 'unipolar', 1: 'bipolar'}
        # polarity_lock = {v:k for k, v in polarity_key.items()}
        #
        # assert polarity in polarity_lock.keys(), f"{polarity} not a valid polarity parameter"
        #
        # {polarity_lock[polarity]}
        pass

    def get_filter(self):
        pass

    def set_filter(self, _filter):
        """
        :param _filter: specifies controlling on unfiltered or filtered readings, True = filtered, False = unfiltered
        :type _filter: bool
        """
        # assert _filter in [True, False], f"{_filter} not valid filter parameter"
        # set_filter = str(int(_filter))
        #
        pass

    def get_delay(self):
        pass

    def set_delay(self, delay):
        """
        :param delay: delay in seconds for setpoint change during autoscanning, 1-255 seconds
        :type delay: int
        """
        # assert delay in range(1, 256), f"{delay} not a valid delay parameter"
        #
        pass


    def set_heater_display(self, display):
        """
        :param display: Display mode for heater. Can either be 'current' or 'power'.
        :type display: string
        """

        if self.output == 2:
            print("Cannot set still heater dispaly")
            return False

        assert display.lower() in heater_display_lock.keys(), f"{display} is not a valid display"

        resp = self.get_heater_setup()
        resp[3] = heater_display_lock[display.lower()]

        self._set_heater_setup(resp)

        self.get_heater_setup()

    # Presumably we're going to know and have set values for heat resistance,
    # max current, etc, maybe that'll simplify this in the future.
    def set_heater_output(self, output, display_type=None):
        """Set heater output with MOUT command.

        :param output: heater output value. If display is 'power', value should
                        be in Watts. If 'current', value should be in percent.
                        If this is the still heater, value should be percent of
                        max voltage (10 V).
        :type output: float
        :param display_type: Display type if you want to set this before setting heater.
                        Can be 'power' or 'current' if output is sample or warm-up heater.
        :type display_type: string

        :returns: heater output
        :rtype: float
        """

        if display_type is not None:
            self.set_heater_display(display_type)

        self.get_heater_range()
        self.get_heater_setup()

        if self.range in ["off", "Off"]:
            print("Heater range is off... Not setting output")
            return False

        if self.output == 0:
            # For sample heater
            max_pow = self.range ** 2 * self.resistance

            if self.display == 'power':
                if 0 <= output <= max_pow:
                    self.ls.msg(f"MOUT {self.output} {output}")
                    return True
                else:
                    print("Cannot set to {} W, max power is {:2e} W".format(
                        output, max_pow))
                    return False

            if self.display == 'current':
                if 0 <= output <= 100:
                    self.ls.msg(f"MOUT {self.output} {output}")
                    return True
                else:
                    print(
                        "Display is current: output must be between 0 and 100")
                    return False

        if self.output == 1:
            # Does anyone actually plan on using the warm-up heater?
            pass

        if self.output == 2:
            if 0 <= output <= 100:
                self.ls.msg(f"MOUT {self.output} {output}")
                return True
            else:
                print("Still heater output must be between 0 and 100%")
                return False

    def get_sample_heater_output(self):
        """Get sample heater output in percent of current or in actual power,
        depending on the heater output selection.

        :return: Sample heater output percentage.

        """
        # Sample Heater
        if self.output == 0:
            resp = self.ls.msg("HTR?")

        return float(resp)

    def get_heater_setup(self):
        """Gets Heater setup params with the HTRSET? command.

        :return resp: List of values that have been returned from the Lakeshore.
        """
        resp = self.ls.msg("HTRSET? {}".format(self.output)).split(',')

        self.resistance = float(resp[0])
        self.max_current = int(resp[1])
        self.max_user_current = float(resp[2].strip('E+'))
        self.display = heater_display_key[resp[3]]

        return resp


    # RAMP, RAMP? - in heater class
    def set_ramp_rate(self, rate):
        pass

    def get_ramp_rate(self, rate):
        pass

    def enable_ramp(self):
        pass

    def disable_ramp(self):
        pass

    # RAMPST?
    def get_ramp_status(self):
        pass

    # RANGE
    def set_heater_range(self, _range):
        """Set heater range with RANGE command.

        :param _range: heater range in amps. Valid values are: "off", "on",
            31.6e-6, 100e-6, 316e-6, 1e-3, 3.16e-3, 10e-3, 31.6e-3, 100e-3
        :type _range: float or str

        :returns: heater range in amps
        :rtype: float
        """
        assert _range in heater_range_lock.keys() or str(_range).lower() in ['on', 'off'], 'Not a valid heater Range'

        if str(_range).lower() == 'off':
            _range = "Off"
        if str(_range).lower() == 'on':
            _range = "On"

        if self.output == 0:
            resp = self.ls.msg(f"RANGE {self.output} {heater_range_lock[_range]}").strip()
        else:
            on_off_key = {"0": "Off", "1": "On"}
            on_off_lock = {v:k for k, v in on_off_key.items()}
            resp = self.ls.msg(f"RANGE {self.output} {on_off_lock[_range]}").strip()

        # refresh self.heater value with RANGE? query
        self.get_heater_range()

    def get_heater_range(self):
        """Get heater range with RANGE? command.

        :returns: heater range in amps
        :rtype: float
        """
        resp = self.ls.msg(f"RANGE? {self.output}").strip()

        if self.output == 0:
            self.range = heater_range_key[resp]
        else:
            on_off_key = {"0": "Off", "1": "On"}
            self.range = on_off_key[resp]

        return self.range

    # SETP - heater class
    def set_setpoint(self, value):
        self.ls.msg(f"SETP {self.output},{value}")

    # SETP? - heater class
    def get_setpoint(self):
        resp = self.ls.msg(f"SETP? {self.output}")
        return resp

    # STILL - heater class?
    def set_still_output(self, value):
        self.ls.msg(f"STILL {value}")

    # STILL? - heater_class?
    def get_still_output(self):
        resp = self.ls.msg(f"STILL?")
        return resp

    # ANALOG, ANALOG?, AOUT?
    # TODO: read up on what analog output is used for, pretty sure just another output
    def get_analog_output(self):
        pass

    def set_analog_output(self):
        pass

    # PID
    def set_pid(self, P, I, D):
        """Set PID parameters for closed loop control.

        :params P: proportional term in PID loop
        :type P: float
        :params I: integral term in PID loop
        :type I: float
        :params D: derivative term in PID loop
        :type D: float

        :returns: response from PID command
        :rtype: str
        """
        assert float(P) <= 1000 and float(P) >= 0
        assert float(I) <= 10000 and float(I) >= 0
        assert float(D) <= 2500 and float(D) >= 0

        resp = self.ls.msg(f"PID {self.output},{P},{I},{D}")
        return resp

    # PID?
    def get_pid(self):
        """Get PID parameters with PID? command.

        :returns: P, I, D
        :rtype: float, float, float
        """
        resp = self.ls.msg("PID?").split(',')
        return float(resp[0]), float(resp[1]), float(resp[2])


if __name__ == "__main__":
    ls = LS372(sys.argv[1])



import serial
import time
import subprocess
import shutil
import pytest
import threading


def create_device_emulator(responses, relay_type='serial'):
    """Create a device emulator fixture.

    This provides a device emulator that can be used to mock a device during
    testing.

    Args:
        responses (dict): Dictionary with commands as keys, and responses as
            values. See :class:`.DeviceEmulator` for details.
        relay_type (str): Currently only 'serial' is implemented. A TCP type
            will be introduced in future versions.

    Returns:
        function:
            A pytest fixture that creates a Device emulator of the specified
            type.

    """
    if relay_type != 'serial':
        raise NotImplementedError(f"relay_type '{relay_type}' is not" +
                                  "implemented or is an invalid type")

    @pytest.fixture()
    def create_device():
        device = DeviceEmulator(responses)
        device.create_serial_relay()

        yield device

        device.shutdown()

    return create_device


class DeviceEmulator:
    """A mocked device which knows how to respond on various communication
    channels.

    Args:
        responses (dict): Initial responses, any response required by Agent
            startup, if any.

    Attributes:
        responses (dict): Current set of responses the DeviceEmulator would give

    """
    def __init__(self, responses):
        self.responses = responses
        self._read = True

    @staticmethod
    def _setup_socat():
        """Setup a data relay with socat.

        The "./responder" link is the external end of the relay, which the Agent
        should connect to. "./internal" is used within the DeviceEmulator object to accept
        commands and to respond to the Agent.

        """
        socat = shutil.which('socat')
        cmd = [socat, '-d', '-d', 'pty,link=./responder,b57600', 'pty,link=./internal,b57600']
        proc = subprocess.Popen(cmd,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE)
        time.sleep(1)

        return proc

    def create_serial_relay(self):
        """Create the serial relay, emulating a hardware device connected over
        serial.

        This first uses ``socat`` to setup a relay. It then connects to the
        "internal" end of the relay, ready to receive communications sent to the
        "responder" end of the relay. This end of the relay is located at
        ``./responder``. You will need to configure your Agent to use that path for
        communication.

        Next it creates a thread to read commands sent to the serial relay in
        the background. This allows responses to be defined within a test using
        DeviceEmulator.define_responses() after instantiation of the DeviceEmulator
        object within a given test.

        """
        self.proc = self._setup_socat()
        self.ser = serial.Serial(
            './internal',
            baudrate=57600,
            timeout=5,
        )
        bkg_read = threading.Thread(name='background', target=self._read_serial)
        bkg_read.start()

    def _read_serial(self):
        """Loop until shutdown, reading any commands sent over the relay.
        Respond immediately to a command with the response in self.responses.

        """
        self._read = True

        while self._read:
            if self.ser.in_waiting > 0:
                msg = self.ser.readline().strip().decode('utf-8')
                print(f"msg={msg}")

                if self.responses is None:
                    continue

                try:
                    if isinstance(self.responses[msg], list):
                        response = self.responses[msg].pop(0)
                    else:
                        response = self.responses[msg]

                    print(f'response={response}')
                    self.ser.write((response + '\r\n').encode('utf-8'))
                except Exception as e:
                    print(f"encountered error {e}")
            time.sleep(0.01)

    def __del__(self):
        self.shutdown()

    def shutdown(self):
        """Shutdown communication on the configured relay. This will stop any
        attempt to read communication on the relay, as well as shutdown the relay
        itself.

        """
        #print('shutting down background reading')
        self._read = False
        time.sleep(1)
        #print('shutting down socat relay')
        self.proc.terminate()
        out, err = self.proc.communicate()
        #print(out, err)

    def create_tcp_relay(self):
        pass

    def define_responses(self, responses):
        """Define what responses are available to reply with on the configured
        communication relay.

        Args:
            responses (dict): Dictionary of commands: response. Values can be a
                list, in which case the responses in the list are popped and given in order
                until depleted.

        Examples:
            The given responses might look like::

                >>> responses = {'KRDG? 1': '+1.7E+03'}
                >>> responses = {'*IDN?': 'LSCI,MODEL425,4250022,1.0',
                                 'RDGFIELD?': ['+1.0E-01', '+1.2E-01', '+1.4E-01']}

        """
        print(f'responses set to {responses}')
        self.responses = responses



from setuptools import setup, find_packages

import versioneer

with open("README.rst", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(name = 'socs',
      long_description=long_description,
      long_description_content_type="text/x-rst",
      version = versioneer.get_version(),
      cmdclass=versioneer.get_cmdclass(),
      description = 'Simons Observatory Control System',
      package_dir = {'socs': 'socs'},
      packages = find_packages(),
      url="https://github.com/simonsobs/socs",
      project_urls={
          "Source Code": "https://github.com/simonsobs/ocs",
          "Documentation": "https://ocs.readthedocs.io/",
          "Bug Tracker": "https://github.com/simonsobs/ocs/issues",
      },
      classifiers=[
          "Programming Language :: Python :: 3",
          "License :: OSI Approved :: BSD License",
          "Intended Audience :: Science/Research",
          "Topic :: Scientific/Engineering :: Astronomy",
          "Framework :: Twisted",
      ],
      python_requires=">=3.6",
      install_requires=[
          'ocs',
          'autobahn[serialization]',
          'twisted',
          'pyserial',
          'sqlalchemy',
          'pysnmp',
      ],
)



import socket


class PrologixInterface:
    def __init__(self, ip_address, gpibAddr, **kwargs):
        self.ip_address = ip_address
        self.gpibAddr = gpibAddr
        self.sock = None
        self.conn_socket()
        self.configure()
        super().__init__(**kwargs)

    def conn_socket(self):
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock.connect((self.ip_address, 1234))
        self.sock.settimeout(5)

    def configure(self):
        self.write('++mode 1\n')
        self.write('++auto 1\n')
        self.write('++addr ' + str(self.gpibAddr))

    def write(self, msg):
        message = msg + '\n'
        self.sock.send(message.encode())

    def read(self):
        return self.sock.recv(128).decode().rstrip('\n').rstrip('\r')

    def version(self):
        self.write('++ver')
        return self.read()

    def identify(self):
        self.write('*idn?')
        return self.read()



"""Michael Randall
    mrandall@ucsd.edu"""

from socs.agent.prologix_interface import PrologixInterface


class TektronixInterface(PrologixInterface):
    def set_freq(self, freq):
        self.write('SOUR:FREQ {:.3f}\n'.format(freq))

    def set_amp(self, amp):
        self.write('SOUR:VOLT {:.3f}\n'.format(amp))

    def set_output(self, state):
        self.write('OUTP:STAT {:.0f}\n'.format(state))
